import os

from ..config import settings
from ..common import progress
from ..common.tools import cprint
from lib_utils import kicad_sym


# KiCad Component Library Manager
class ComponentLibManager(object):
    def __init__(self, library_path):
        # Load library and template paths
        cprint(f'[KCAD]\tlibrary_path: {library_path}', silent=settings.SILENT)

        # Check files exist
        if not os.path.isfile(library_path):
            cprint(f'[KCAD]\tError loading library file ({library_path})', silent=settings.SILENT)
            return None

        # Load library
        self.kicad_lib = kicad_sym.KicadLibrary.from_file(library_path)
        self.library_name = library_path.split(os.sep)[-1]
        cprint('[KCAD]\tNumber of parts in library ' + self.library_name + ': ' + str(len(self.kicad_lib.symbols)), silent=settings.SILENT)

    def is_symbol_in_library(self, symbol_part_number):
        ''' Check if symbol already exists in library '''
        for symbol in self.kicad_lib.symbols:
            cprint(f'[DBUG]\t{symbol.name} ?= {symbol_part_number}', silent=settings.HIDE_DEBUG)
            if symbol.name == symbol_part_number:
                cprint(f'[KCAD]\tWarning: Component {symbol_part_number} already in library', silent=settings.SILENT)
                return True

        return False

    def add_symbol_to_library_from_inventree(self, symbol_data, template_path=None, show_progress=True):
        ''' Create symbol in KiCad library '''
        part_in_lib = False
        new_part = False
        category = symbol_data['category'][0]
        subcategory = symbol_data['category'][1]

        if not template_path:
            # Fetch template path
            try:
                template_path = settings.symbol_templates_paths[category][subcategory]
            except:
                template_path = settings.symbol_templates_paths[category]['Default']

        # Check files exist
        if not self.kicad_lib:
            return part_in_lib, new_part
        if not os.path.isfile(template_path):
            cprint(f'[KCAD]\tError loading template file ({template_path})', silent=settings.SILENT)
            return part_in_lib, new_part

        # Check if part already in library
        try:
            is_symbol_in_library = self.is_symbol_in_library(symbol_data['IPN'])
            part_in_lib = True
        except:
            is_symbol_in_library = False
        if is_symbol_in_library:
            return part_in_lib, new_part

        # Progress Update
        if show_progress and not progress.update_progress_bar_window():
            return part_in_lib, new_part

        # Load template
        templatelib = kicad_sym.KicadLibrary.from_file(template_path)
        # Load new symbol
        if len(templatelib.symbols) == 1:
            for symbol in templatelib.symbols:
                new_symbol = symbol
        else:
            cprint('[KCAD]\tError: Found more than 1 symbol template in template file, aborting', silent=settings.SILENT)
            return part_in_lib, new_part

        # Update name
        new_symbol.name = symbol_data['IPN']

        # Update properties
        for property in new_symbol.properties:
            # Main data
            if property.value in symbol_data.keys():
                property.value = symbol_data[property.value]
                continue

            # Parameters
            if property.value in symbol_data['parameters'].keys():
                property.value = symbol_data['parameters'][property.value]
                continue

            # Special properties
            if property.name in ['Value', 'Manufacturer', 'Manufacturer Part Number']:
                if property.name == 'Value':
                    property.value = symbol_data['IPN']
                elif property.name == 'Manufacturer':
                    property.value = list(symbol_data['manufacturer'].keys())[0]
                elif property.name == 'Manufacturer Part Number':
                    property.value = list(symbol_data['manufacturer'].values())[0][0]
                continue

        # Add symbol to library
        self.kicad_lib.symbols.append(new_symbol)
        # Update generator version
        self.kicad_lib.version = '20211014'
        # Write library
        self.kicad_lib.write()

        cprint(f'[KCAD]\tSuccess: Component added to library {self.library_name}', silent=settings.SILENT)
        part_in_lib = True
        new_part = True

        # Progress Update
        if show_progress and not progress.update_progress_bar_window():
            pass

        return part_in_lib, new_part

    # NOT YET SUPPORTED - REMOVE?
    # def delete_symbol_from_lib(self, part_number):
    #     ''' Remove symbol from KiCad library '''
    #     pass



from . import kicad_symbol


def inventree_to_kicad(part_data: dict, library_path: str, template_path=None, show_progress=True) -> bool:
    ''' Create KiCad symbol from InvenTree part data '''
    klib = kicad_symbol.ComponentLibManager(library_path)
    return klib.add_symbol_to_library_from_inventree(symbol_data=part_data,
                                                     template_path=template_path,
                                                     show_progress=show_progress)

# NOT YET SUPPORTED - REMOVE?
# def delete_part(part_number: str, library_path: str) -> bool:
#     ''' Delete KiCad symbol from library '''
#     pass



import copy
import os

from ..config import settings
from ..common import part_tools, progress
from ..common.tools import cprint
from ..config import config_interface
from ..database import inventree_api
from fuzzywuzzy import fuzz
from ..search import search_api, digikey_api, mouser_api, element14_api, lcsc_api


def connect_to_server(timeout=5) -> bool:
    ''' Connect to InvenTree server using user settings '''
    connect = False
    settings.load_inventree_settings()

    try:
        connect = inventree_api.connect(server=settings.SERVER_ADDRESS,
                                        username=settings.USERNAME,
                                        password=settings.PASSWORD,
                                        connect_timeout=timeout)
    except TimeoutError:
        pass

    if not connect:
        if not settings.SERVER_ADDRESS:
            cprint('[TREE]\tError connecting to InvenTree server: missing server address')
            return connect
        if not settings.USERNAME:
            cprint('[TREE]\tError connecting to InvenTree server: missing username')
            return connect
        if not settings.PASSWORD:
            cprint('[TREE]\tError connecting to InvenTree server: missing password')
            return connect
        cprint('[TREE]\tError connecting to InvenTree server: invalid address, username or password')
    else:
        env = [env_type.name for env_type in settings.Environment
               if env_type.value == settings.environment][0]
        cprint(f'[TREE]\tSuccessfully connected to InvenTree server (ENV={env})', silent=settings.SILENT)

    return connect


def get_categories(part_info: dict, supplier_only=False) -> list:
    ''' Find categories from part supplier data, use "somewhat automatic" matching '''
    categories = [None, None]

    try:
        supplier_category = str(part_info['category'])
        supplier_subcategory = str(part_info['subcategory'])
    except KeyError:
        return categories

    # Return supplier category, if match not needed
    if supplier_only:
        categories[0] = supplier_category
        categories[1] = supplier_subcategory
        return categories

    function_filter = False
    # TODO: Make 'filter_parameter' user defined?
    filter_parameter = 'Function Type'

    # Check existing matches
    # Load inversed category map
    category_map = config_interface.load_supplier_categories_inversed(supplier_config_path=settings.CONFIG_DIGIKEY_CATEGORIES)

    try:
        for inventree_category in category_map.keys():
            for key, inventree_subcategory in category_map[inventree_category].items():
                if supplier_subcategory == key:
                    categories[0] = inventree_category
                    # Check if filtering by function
                    if inventree_subcategory.startswith(config_interface.FUNCTION_FILTER_KEY):
                        function_filter = True

                    # Save subcategory if not function filtered
                    if not function_filter:
                        categories[1] = inventree_subcategory

                    break
    except:
        pass

    # Function Filter
    if not categories[1] and function_filter:
        cprint(f'[INFO]\tSubcategory is filtered using "{filter_parameter}" parameter', silent=settings.SILENT, end='')
        # Load parameter map
        parameter_map = config_interface.load_category_parameters(categories[0], settings.CONFIG_SUPPLIER_PARAMETERS)
        # Build compare list
        compare = []
        for supplier_parameter, inventree_parameter in parameter_map.items():
            if (supplier_parameter in part_info['parameters'].keys() and inventree_parameter == filter_parameter):
                compare.append(part_info['parameters'][supplier_parameter])

        # Load subcategory map
        category_map = config_interface.load_supplier_categories(supplier_config_path=settings.CONFIG_DIGIKEY_CATEGORIES)[categories[0]]
        for inventree_subcategory in category_map.keys():
            for item in compare:
                fuzzy_match = fuzz.partial_ratio(inventree_subcategory, item)
                display_result = f'"{inventree_subcategory}" ?= "{item}"'.ljust(50)
                cprint(f'{display_result} => {fuzzy_match}', silent=settings.HIDE_DEBUG)
                if fuzzy_match >= settings.CATEGORY_MATCH_RATIO_LIMIT:
                    categories[1] = inventree_subcategory.replace(config_interface.FUNCTION_FILTER_KEY, '')
                    break

            if categories[1]:
                cprint('\t[ PASS ]', silent=settings.SILENT)
                break

    if not categories[1] and function_filter:
        cprint('\t[ FAILED ]', silent=settings.SILENT)

    # Automatic Match
    if not (categories[0] and categories[1]):
        # Load category map
        category_map = config_interface.load_supplier_categories(supplier_config_path=settings.CONFIG_DIGIKEY_CATEGORIES)

        def find_supplier_category_match(supplier_category: str, ignore_categories=False):
            # Check for match with Inventree categories
            category_match = None
            subcategory_match = None

            for inventree_category in category_map.keys():
                fuzzy_match = 0
                
                if not ignore_categories:
                    fuzzy_match = fuzz.partial_ratio(supplier_category, inventree_category)
                    display_result = f'"{supplier_category}" ?= "{inventree_category}"'.ljust(50)
                    cprint(f'{display_result} => {fuzzy_match}', silent=settings.HIDE_DEBUG)

                if fuzzy_match < settings.CATEGORY_MATCH_RATIO_LIMIT and category_map[inventree_category]:
                    # Compare to subcategories
                    for inventree_subcategory in category_map[inventree_category]:
                        fuzzy_match = fuzz.partial_ratio(supplier_category, inventree_subcategory)
                        display_result = f'"{supplier_category}" ?= "{inventree_subcategory}"'.ljust(50)
                        cprint(f'{display_result} => {fuzzy_match}', silent=settings.HIDE_DEBUG)

                        if fuzzy_match >= settings.CATEGORY_MATCH_RATIO_LIMIT:
                            subcategory_match = inventree_subcategory
                            break

                if fuzzy_match >= settings.CATEGORY_MATCH_RATIO_LIMIT:
                    category_match = inventree_category
                    break

            return category_match, subcategory_match

        # Find category and subcategories match
        category, subcategory = find_supplier_category_match(supplier_category)
        if category:
            categories[0] = category
        if subcategory:
            categories[1] = subcategory

        # Run match with supplier subcategory
        if not categories[0] or not categories[1]:
            if categories[0]:
                # If category was found: ignore them for the comparison
                category, subcategory = find_supplier_category_match(supplier_subcategory, ignore_categories=True)
            else:
                category, subcategory = find_supplier_category_match(supplier_subcategory)

        if category and not categories[0]:
            categories[0] = category
        if subcategory and not categories[1]:
            categories[1] = subcategory

    # Final checks
    if not categories[0]:
        cprint(f'[INFO]\tWarning: "{part_info["category"]}" did not match any supplier category ', silent=settings.SILENT)
    else:
        cprint(f'[INFO]\tCategory: "{categories[0]}"', silent=settings.SILENT)
    if not categories[1]:
        cprint(f'[INFO]\tWarning: "{part_info["subcategory"]}" did not match any supplier subcategory ', silent=settings.SILENT)
    else:
        cprint(f'[INFO]\tSubcategory: "{categories[1]}"', silent=settings.SILENT)

    return categories


def translate_form_to_inventree(part_info: dict, categories: list, is_custom=False) -> dict:
    ''' Using supplier part data and categories, fill-in InvenTree part dictionary '''

    # Copy template
    inventree_part = copy.deepcopy(settings.inventree_part_template)

    # Translate form data to inventree part
    inventree_part['category'][0] = categories[0]
    inventree_part['category'][1] = categories[1]
    inventree_part['name'] = part_info['name']
    inventree_part['description'] = part_info['description']
    inventree_part['revision'] = part_info['revision']
    inventree_part['keywords'] = part_info['keywords']
    inventree_part['supplier'] = {
        part_info['supplier_name']: [
            part_info['supplier_part_number'],
        ],
    }
    inventree_part['manufacturer'] = {
        part_info['manufacturer_name']: [
            part_info['manufacturer_part_number'],
        ],
    }
    # Replace whitespaces in URL
    inventree_part['supplier_link'] = part_info['supplier_link'].replace(' ', '%20')
    inventree_part['datasheet'] = part_info['datasheet'].replace(' ', '%20')
    # Image URL is not shown to user so force default key/value
    inventree_part['image'] = part_info['image'].replace(' ', '%20')

    # Load parameters map
    parameter_map = config_interface.load_category_parameters(category=inventree_part["category"][0],
                                                              supplier_config_path=settings.CONFIG_SUPPLIER_PARAMETERS, )

    if not is_custom:
        # Add Parameters
        if parameter_map:
            for supplier_param, inventree_param in parameter_map.items():
                # Some parameters may not be mapped
                if inventree_param not in inventree_part['parameters'].keys():
                    if supplier_param != 'Manufacturer Part Number':
                        try:
                            parameter_value = part_tools.clean_parameter_value(category=categories[0],
                                                                               name=supplier_param,
                                                                               value=part_info['parameters'][supplier_param])
                            inventree_part['parameters'][inventree_param] = parameter_value
                        except:
                            cprint(f'[INFO]\tWarning: Parameter "{supplier_param}" not found in supplier data', silent=settings.SILENT)
                    else:
                        inventree_part['parameters'][inventree_param] = part_info['manufacturer_part_number']

            # Check for missing parameters and fill value with dash
            for inventree_param in parameter_map.values():
                if inventree_param not in inventree_part['parameters'].keys():
                    inventree_part['parameters'][inventree_param] = '-'
        else:
            cprint(f'[INFO]\tWarning: Parameter map for "{categories[0]}" does not exist or is empty', silent=settings.SILENT)

    return inventree_part


def translate_supplier_to_form(supplier: str, part_info: dict) -> dict:
    ''' Translate supplier data to user form format '''

    part_form = {}

    def get_value_from_user_key(user_key: str, default_key: str, default_value=None) -> str:
        ''' Get value mapped from user search key, else default search key '''
        user_search_key = None
        if supplier == 'Digi-Key':
            user_search_key = settings.CONFIG_DIGIKEY.get(user_key, None)
        elif supplier == 'Mouser':
            user_search_key = settings.CONFIG_MOUSER.get(user_key, None)
        elif supplier in ['Farnell', 'Newark', 'Element14']:
            user_search_key = settings.CONFIG_ELEMENT14.get(user_key, None)
        elif supplier == 'LCSC':
            user_search_key = settings.CONFIG_LCSC.get(user_key, None)
        else:
            return default_value
        
        # If no user key, use default
        if not user_search_key:
            return part_info.get(default_key, default_value)

        # Get value for user key, return value from default key if not found
        return part_info.get(user_search_key, part_info.get(default_key, default_value))

    def get_supplier_name(supplier: str) -> str:
        ''' Get InvenTree supplier name '''
        # Check that supplier is supported
        if supplier not in settings.SUPPORTED_SUPPLIERS_API:
            return ''

        if supplier == 'Digi-Key':
            supplier_name = settings.CONFIG_DIGIKEY.get('SUPPLIER_INVENTREE_NAME', None)
        elif supplier == 'Mouser':
            supplier_name = settings.CONFIG_MOUSER.get('SUPPLIER_INVENTREE_NAME', None)
        # elif supplier in ['Farnell', 'Newark', 'Element14']:
        #     supplier_name = settings.CONFIG_ELEMENT14.get('SUPPLIER_INVENTREE_NAME', None)
        elif supplier == 'LCSC':
            supplier_name = settings.CONFIG_LCSC.get('SUPPLIER_INVENTREE_NAME', None)
        else:
            supplier_name = supplier
        
        return supplier_name

    # Check that supplier argument is valid
    if not supplier or (supplier != 'custom' and supplier not in settings.SUPPORTED_SUPPLIERS_API):
        return part_form
    # Get default keys
    if supplier == 'Digi-Key':
        default_search_keys = digikey_api.get_default_search_keys()
    elif supplier == 'Mouser':
        default_search_keys = mouser_api.get_default_search_keys()
    elif supplier in ['Farnell', 'Newark', 'Element14']:
        default_search_keys = element14_api.get_default_search_keys()
    elif supplier == 'LCSC':
        default_search_keys = lcsc_api.get_default_search_keys()
    else:
        # Empty array of default search keys
        default_search_keys = [''] * len(digikey_api.get_default_search_keys())

    # Translate supplier data to form fields
    part_form['name'] = get_value_from_user_key('SEARCH_NAME', default_search_keys[0], default_value='')
    part_form['description'] = get_value_from_user_key('SEARCH_DESCRIPTION', default_search_keys[1], default_value='')
    part_form['revision'] = get_value_from_user_key('SEARCH_REVISION', default_search_keys[2], default_value=settings.INVENTREE_DEFAULT_REV)
    part_form['keywords'] = get_value_from_user_key('SEARCH_KEYWORDS', default_search_keys[1], default_value='')
    part_form['supplier_name'] = get_supplier_name(supplier)
    part_form['supplier_part_number'] = get_value_from_user_key('SEARCH_SKU', default_search_keys[4], default_value='')
    part_form['supplier_link'] = get_value_from_user_key('SEARCH_SUPPLIER_URL', default_search_keys[7], default_value='')
    part_form['manufacturer_name'] = get_value_from_user_key('SEARCH_MANUFACTURER', default_search_keys[5], default_value='')
    part_form['manufacturer_part_number'] = get_value_from_user_key('SEARCH_MPN', default_search_keys[6], default_value='')
    part_form['datasheet'] = get_value_from_user_key('SEARCH_DATASHEET', default_search_keys[8], default_value='')
    part_form['image'] = get_value_from_user_key('', default_search_keys[9], default_value='')
    
    return part_form


def supplier_search(supplier: str, part_number: str, test_mode=False) -> dict:
    ''' Wrapper for supplier search, allow use of cached data (limited daily API calls) '''
    part_info = {}
    # Check part number exist
    if not part_number:
        cprint('\n[MAIN]\tError: Missing Part Number', silent=settings.SILENT)
        return part_info

    # Load from file if cache is enabled
    store = ''
    if supplier in ['Farnell', 'Newark', 'Element14']:
        element14_config = config_interface.load_file(settings.CONFIG_ELEMENT14_API)
        store = element14_config.get(f'{supplier.upper()}_STORE', '').replace(' ', '')
    search_filename = settings.search_results['directory'] + supplier + store + '_' + part_number + settings.search_results['extension']

    # Get cached data
    part_info = search_api.load_from_file(search_filename, test_mode)
    if part_info:
        cprint(f'\n[MAIN]\tUsing {supplier} cached data for {part_number}', silent=settings.SILENT)
    else:
        cprint(f'\n[MAIN]\t{supplier} search for {part_number}', silent=settings.SILENT)
        if supplier == 'Digi-Key':
            part_info = digikey_api.fetch_part_info(part_number)
        elif supplier == 'Mouser':
            part_info = mouser_api.fetch_part_info(part_number)
        elif supplier in ['Farnell', 'Newark', 'Element14']:
            part_info = element14_api.fetch_part_info(part_number, supplier)
        elif supplier == 'LCSC':
            part_info = lcsc_api.fetch_part_info(part_number)

    # Check supplier data exist
    if not part_info:
        cprint(f'[INFO]\tError: Failed to fetch data for "{part_number}"', silent=settings.SILENT)

    # Save search results
    if part_info:
        search_api.save_to_file(part_info, search_filename)

    return part_info


def inventree_fuzzy_company_match(name: str) -> str:
    ''' Fuzzy match company name to exisiting companies '''
    inventree_companies = inventree_api.get_all_companies()

    for company_name in inventree_companies.keys():
        cprint(f'{name.lower()} == {company_name.lower()} % {fuzz.partial_ratio(name.lower(), company_name.lower())}',
               silent=settings.HIDE_DEBUG)
        if fuzz.partial_ratio(name.lower(), company_name.lower()) == 100:
            return company_name
    
    return name


def inventree_create_manufacturer_part(part_id: int, manufacturer_name: str, manufacturer_mpn: str, datasheet: str, description: str) -> bool:
    ''' Create manufacturer part '''

    cprint('\n[MAIN]\tCreating manufacturer part', silent=settings.SILENT)
    manufacturer_part = inventree_api.is_new_manufacturer_part(manufacturer_name=manufacturer_name,
                                                               manufacturer_mpn=manufacturer_mpn)

    if manufacturer_part:
        cprint('[INFO]\tManufacturer part already exists, skipping.', silent=settings.SILENT)
    else:
        # Create a new manufacturer part
        is_manufacturer_part_created = inventree_api.create_manufacturer_part(part_id=part_id,
                                                                              manufacturer_name=manufacturer_name,
                                                                              manufacturer_mpn=manufacturer_mpn,
                                                                              datasheet=datasheet,
                                                                              description=description)

        if is_manufacturer_part_created:
            cprint('[INFO]\tSuccess: Added new manufacturer part', silent=settings.SILENT)
            return True

    return False


def inventree_create_supplier_part(part) -> bool:
    return


def inventree_create(part_info: dict, categories: list, kicad=False, symbol=None, footprint=None, show_progress=True, is_custom=False):
    ''' Create InvenTree part from supplier part data and categories '''

    part_pk = 0
    new_part = False

    # Translate to InvenTree part format
    inventree_part = translate_form_to_inventree(part_info=part_info,
                                                 categories=categories,
                                                 is_custom=is_custom)

    if not inventree_part:
        cprint('\n[MAIN]\tError: Failed to process form data', silent=settings.SILENT)

    # Fetch category info from InvenTree part
    category_name = inventree_part['category'][0]
    subcategory_name = inventree_part['category'][1]
    category_pk = inventree_api.get_inventree_category_id(category_name)
    category_select = category_pk

    # Check if subcategory exists
    if subcategory_name:
        # Fetch subcategory id
        subcategory_pk = inventree_api.get_inventree_category_id(category_name=subcategory_name,
                                                                 parent_category_id=category_pk)
        
        if subcategory_pk <= 0:
            cprint(f'\n[TREE]\tWarning: Subcategory "{subcategory_name}" does not exist', silent=settings.SILENT)
            
            # Check if user enabled option to automatically create the subcategory in general settings
            if settings.AUTOMATIC_SUBCATEGORY_CREATE:
                subcategory_pk, is_subcategory_new = inventree_api.create_category(parent=category_name, name=subcategory_name)
                if subcategory_pk > 0:
                    cprint(f'[TREE]\tSuccess: Subcategory "{category_name}/{subcategory_name}" was automatically added to InvenTree')

        if subcategory_pk > 0:
            category_select = subcategory_pk

    # Progress Update
    if show_progress and not progress.update_progress_bar_window():
        return new_part, part_pk, inventree_part

    if category_select <= 0:
        cprint(f'[ERROR]\tCategories ({category_name} - {subcategory_name}) does not exist in InvenTree', silent=settings.SILENT)
    else:
        # Check if part already exists
        part_pk = inventree_api.is_new_part(category_select, inventree_part)
        # Part exists
        if part_pk > 0:
            cprint('[INFO]\tPart already exists, skipping.', silent=settings.SILENT)
            ipn = inventree_api.get_part_number(part_pk)
            if ipn:
                # Update InvenTree part number
                inventree_part['IPN'] = ipn
                # Update InvenTree URL
                inventree_part['inventree_url'] = f'{settings.PART_URL_ROOT}{inventree_part["IPN"]}/'
            else:
                inventree_part['inventree_url'] = f'{settings.PART_URL_ROOT}{part_pk}/'
        # Part is new
        else:
            new_part = True
            # Create a new Part
            # Use the pk (primary-key) of the category
            part_pk = inventree_api.create_part(category_id=category_select,
                                                name=inventree_part['name'],
                                                description=inventree_part['description'],
                                                revision=inventree_part['revision'],
                                                image=inventree_part['image'],
                                                keywords=inventree_part['keywords'])

            # Check part primary key
            if not part_pk:
                return new_part, part_pk, inventree_part
            # Progress Update
            if show_progress and not progress.update_progress_bar_window():
                return new_part, part_pk, inventree_part

            if settings.CONFIG_IPN.get('IPN_ENABLE_CREATE', True):
                # Generate Internal Part Number
                cprint('\n[MAIN]\tGenerating Internal Part Number', silent=settings.SILENT)
                ipn = part_tools.generate_part_number(category_name, part_pk)
                cprint(f'[INFO]\tInternal Part Number = {ipn}', silent=settings.SILENT)
                # Update InvenTree part number
                ipn_update = inventree_api.set_part_number(part_pk, ipn)
                if not ipn_update:
                    cprint('\n[INFO]\tError updating IPN', silent=settings.SILENT)
                inventree_part['IPN'] = ipn
                # Update InvenTree URL
                inventree_part['inventree_url'] = f'{settings.PART_URL_ROOT}{inventree_part["IPN"]}/'
            else:
                inventree_part['inventree_url'] = f'{settings.PART_URL_ROOT}{part_pk}/'

    # Progress Update
    if show_progress and not progress.update_progress_bar_window():
        return new_part, part_pk, inventree_part

    if part_pk > 0:
        if new_part:
            cprint('[INFO]\tSuccess: Added new part to InvenTree', silent=settings.SILENT)
            if inventree_part['image']:
                # Add image
                image_result = inventree_api.upload_part_image(inventree_part['image'], part_pk)
                if not image_result:
                    cprint('[TREE]\tWarning: Failed to upload part image', silent=settings.SILENT)

        if kicad:
            # Update IPN if it does not exist
            if not inventree_part['IPN']:
                ipn = list(inventree_part['manufacturer'].values())[0][0]
                inventree_part['IPN'] = ipn

            # Create mandatory parameters (symbol & footprint)
            if symbol and ipn:
                kicad_symbol = symbol + ':' + ipn
            else:
                try:
                    kicad_symbol = settings.symbol_libraries_paths[category_name].split(os.sep)[-1].split('.')[0] + ':' + ipn
                except:
                    kicad_symbol = category_name.replace(' ', '_') + ':TBD'
            if footprint:
                kicad_footprint = footprint
            else:
                try:
                    supplier_package = inventree_part['parameters']['Package Type']
                    kicad_footprint = settings.footprint_lookup_table[category_name][supplier_package]
                except:
                    kicad_footprint = category_name.replace(' ', '_') + ':TBD'

            # Add symbol & footprint to InvenTree part
            inventree_part['parameters']['Symbol'] = kicad_symbol
            inventree_part['parameters']['Footprint'] = kicad_footprint

        if not inventree_part['parameters']:
            category_parameters = inventree_api.get_category_parameters(category_select)

            # Add category-defined parameters
            for parameter in category_parameters:
                inventree_part['parameters'][parameter[0]] = parameter[1]

        # Create parameters
        if len(inventree_part['parameters']) > 0:
            cprint('\n[MAIN]\tCreating parameters', silent=settings.SILENT)
            parameters_lists = [
                [],  # Store new parameters
                [],  # Store existings parameters
            ]
            for name, value in inventree_part['parameters'].items():
                parameter, is_new_parameter = inventree_api.create_parameter(part_id=part_pk, template_name=name, value=value)
                # Progress Update
                if show_progress and not progress.update_progress_bar_window(3):
                    return new_part, part_pk, inventree_part

                if is_new_parameter:
                    parameters_lists[0].append(name)
                else:
                    parameters_lists[1].append(name)

            if parameters_lists[0]:
                cprint('[INFO]\tSuccess: The following parameters were created:', silent=settings.SILENT)
                for item in parameters_lists[0]:
                    cprint(f'--->\t{item}', silent=settings.SILENT)
            if parameters_lists[1]:
                cprint('[TREE]\tWarning: The following parameters were skipped:', silent=settings.SILENT)
                for item in parameters_lists[1]:
                    cprint(f'--->\t{item}', silent=settings.SILENT)

        # Create manufacturer part(s)
        # TODO: Support multiple MPNs?
        for manufacturer_name, manufacturer_mpns in inventree_part['manufacturer'].items():
            # Overwrite manufacturer name with matching one from database
            manufacturer_name = inventree_fuzzy_company_match(manufacturer_name)
            # Get MPN
            manufacturer_mpn = manufacturer_mpns[0]

            if manufacturer_mpn:
                cprint('\n[MAIN]\tCreating manufacturer part', silent=settings.SILENT)
                manufacturer_part = inventree_api.is_new_manufacturer_part(manufacturer_name=manufacturer_name,
                                                                           manufacturer_mpn=manufacturer_mpn)

                if manufacturer_part:
                    cprint('[INFO]\tManufacturer part already exists, skipping.', silent=settings.SILENT)
                else:
                    # Create a new manufacturer part
                    is_manufacturer_part_created = inventree_api.create_manufacturer_part(part_id=part_pk,
                                                                                          manufacturer_name=manufacturer_name,
                                                                                          manufacturer_mpn=manufacturer_mpn,
                                                                                          datasheet=inventree_part['datasheet'],
                                                                                          description=inventree_part['description'])

                    if is_manufacturer_part_created:
                        cprint('[INFO]\tSuccess: Added new manufacturer part', silent=settings.SILENT)

        # Create supplier part(s)
        # TODO: Support multiple SKUs?
        for supplier_name, supplier_skus in inventree_part['supplier'].items():
            # Get SKU
            supplier_sku = supplier_skus[0]

            if supplier_sku:
                cprint('\n[MAIN]\tCreating supplier part', silent=settings.SILENT)
                is_new_supplier_part = inventree_api.is_new_supplier_part(supplier_name=supplier_name,
                                                                          supplier_sku=supplier_sku)

                if not is_new_supplier_part:
                    cprint('[INFO]\tSupplier part already exists, skipping.', silent=settings.SILENT)
                else:
                    # Create a new supplier part
                    is_supplier_part_created = inventree_api.create_supplier_part(part_id=part_pk,
                                                                                  manufacturer_name=manufacturer_name,
                                                                                  manufacturer_mpn=manufacturer_mpn,
                                                                                  supplier_name=supplier_name,
                                                                                  supplier_sku=supplier_sku,
                                                                                  description=inventree_part['description'],
                                                                                  link=inventree_part['supplier_link'])

                    if is_supplier_part_created:
                        cprint('[INFO]\tSuccess: Added new supplier part', silent=settings.SILENT)

    # Progress Update
    if show_progress and not progress.update_progress_bar_window(3):
        pass

    return new_part, part_pk, inventree_part


def inventree_create_alternate(part_info: dict, part_id='', part_ipn='', show_progress=True) -> bool:
    ''' Create alternate manufacturer and supplier entries for an existing InvenTree part '''

    result = False
    cprint('\n[MAIN]\tSearching for original part in database', silent=settings.SILENT)
    part = inventree_api.fetch_part(part_id, part_ipn)

    # Progress Update
    if show_progress and not progress.update_progress_bar_window(3):
        return

    if part:
        part_pk = part.pk
        part_description = part.description
        cprint(f'[INFO] Success: Found original part in database (ID = {part_pk} | Description = "{part_description}")', silent=settings.SILENT)
    else:
        cprint('[INFO] Error: Original part was not found in database', silent=settings.SILENT)
        return result

    # Overwrite manufacturer name with matching one from database
    manufacturer_name = inventree_fuzzy_company_match(part_info.get('manufacturer_name', ''))
    manufacturer_mpn = part_info.get('manufacturer_part_number', '')
    datasheet = part_info.get('datasheet', '')

    # Create manufacturer part
    if manufacturer_mpn:
        inventree_create_manufacturer_part(part_id=part_pk,
                                           manufacturer_name=manufacturer_name,
                                           manufacturer_mpn=manufacturer_mpn,
                                           datasheet=datasheet,
                                           description=part_description)

    # Progress Update
    if show_progress and not progress.update_progress_bar_window(3):
        return

    supplier_name = part_info.get('supplier_name', '')
    supplier_sku = part_info.get('supplier_part_number', '')
    supplier_link = part_info.get('supplier_link', '')

    # Add supplier alternate
    if supplier_sku:
        cprint('\n[MAIN]\tCreating supplier part', silent=settings.SILENT)
        is_new_supplier_part = inventree_api.is_new_supplier_part(supplier_name=supplier_name,
                                                                  supplier_sku=supplier_sku)

        if not is_new_supplier_part:
            cprint('[INFO]\tSupplier part already exists, skipping.', silent=settings.SILENT)
        else:
            # Create a new supplier part
            is_supplier_part_created = inventree_api.create_supplier_part(part_id=part_pk,
                                                                          manufacturer_name=manufacturer_name,
                                                                          manufacturer_mpn=manufacturer_mpn,
                                                                          supplier_name=supplier_name,
                                                                          supplier_sku=supplier_sku,
                                                                          description=part_description,
                                                                          link=supplier_link)

            if is_supplier_part_created:
                cprint('[INFO]\tSuccess: Added new supplier part', silent=settings.SILENT)
                result = True

    return result



import sys

from .config import settings
from .common.tools import cprint
from .config import config_interface
from .database import inventree_api, inventree_interface


def setup_inventree():
    SETUP_CATEGORIES = True
    SETUP_PARAMETERS = True

    if SETUP_CATEGORIES or SETUP_PARAMETERS:
        cprint('\n[MAIN]\tStarting InvenTree setup', silent=settings.SILENT)
        # Load category configuration file
        categories = config_interface.load_file(settings.CONFIG_CATEGORIES)['CATEGORIES']
        # cprint(categories)

        cprint('[MAIN]\tConnecting to Inventree', silent=settings.SILENT)
        inventree_connect = inventree_interface.connect_to_server()

        if not inventree_connect:
            sys.exit(-1)

    if SETUP_CATEGORIES:
        for category in categories.keys():
            cprint(f'\n[MAIN]\tCreating {category.upper()}')
            category_pk, is_category_new = inventree_api.create_category(parent=None, name=category)
            if is_category_new:
                cprint(f'[TREE]\tSuccess: Category "{category}" was added to InvenTree')
            else:
                cprint(f'[TREE]\tWarning: Category "{category}" already exists')

            if categories[category]:
                cprint('[MAIN]\tCreating Subcategories')
                for subcategory in categories[category]:
                    sub_category_pk, is_subcategory_new = inventree_api.create_category(parent=category, name=subcategory)

                    if is_subcategory_new:
                        cprint(f'[TREE]\tSuccess: Subcategory "{category}/{subcategory}" was added to InvenTree')
                    else:
                        cprint(f'[TREE]\tWarning: Subcategory "{category}/{subcategory}" already exists')

    if SETUP_PARAMETERS:
        # Load parameter configuration file
        parameters = config_interface.load_file(settings.CONFIG_PARAMETERS)
        # cprint(parameters)
        cprint('\n[MAIN]\tLoading Parameters')
        for name, unit in parameters.items():
            pk = inventree_api.create_parameter_template(name, unit)
            if pk > 0:
                cprint(f'[TREE]\tSuccess: Parameter "{name}" was added to InvenTree')
            else:
                cprint(f'[TREE]\tWarning: Parameter "{name}" already exists')


if __name__ == '__main__':
    setup_inventree()



import os
import sys
import platform
from enum import Enum

from ..common.tools import cprint
from .import config_interface

# DEBUG
# Testing
ENABLE_TEST = False
# Silent Mode
SILENT = False
# Debug
HIDE_DEBUG = True


def enable_test_mode():
    global ENABLE_TEST
    global SILENT
    ENABLE_TEST = True
    SILENT = True


# PATHS
if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
    PROJECT_DIR = os.path.abspath(os.path.dirname(sys.executable))
else:
    PROJECT_DIR = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
# InvenTree API
sys.path.append(os.path.join(PROJECT_DIR, 'database', 'inventree-python'))
# Digi-Key API
sys.path.append(os.path.join(PROJECT_DIR, 'search', 'digikey_api'))
# KiCad Library Utils
sys.path.append(os.path.join(PROJECT_DIR, 'kicad'))
# Tests
sys.path.append(os.path.join(PROJECT_DIR, 'tests'))

# HOME FOLDER
USER_HOME = os.path.expanduser("~")
# APP NAME
APP_NAME = 'kintree'
# CONFIG PATH
if platform.system() == 'Linux':
    HOME_DIR = os.path.join(USER_HOME, '.config', APP_NAME, '')
else:
    HOME_DIR = os.path.join(USER_HOME, APP_NAME, '')
# Create config path if it does not exists
if not os.path.exists(HOME_DIR):
    os.makedirs(HOME_DIR, exist_ok=True)


# USER AND CONFIG FILES
def load_user_config():
    global USER_SETTINGS
    global CONFIG_ROOT
    global CONFIG_USER_FILES

    USER_SETTINGS = config_interface.load_user_paths(home_dir=HOME_DIR)
    CONFIG_ROOT = os.path.join(PROJECT_DIR, 'config', '')
    CONFIG_USER_FILES = os.path.join(USER_SETTINGS['USER_FILES'], '')

    # Create user files folder if it does not exists
    if not os.path.exists(CONFIG_USER_FILES):
        os.makedirs(CONFIG_USER_FILES)
    # Create user files
    return config_interface.load_user_config_files(path_to_root=CONFIG_ROOT,
                                                   path_to_user_files=CONFIG_USER_FILES,
                                                   silent=HIDE_DEBUG)


# Load user config
if not load_user_config():
    # Check if configuration files already exist
    if not os.path.isfile(os.path.join(CONFIG_USER_FILES, 'categories.yaml')):
        cprint('\n[ERROR]\tSome Ki-nTree configuration files seem to be missing')
        exit(-1)

# KiCad
CONFIG_KICAD = os.path.join(CONFIG_USER_FILES, 'kicad.yaml')
CONFIG_KICAD_CATEGORY_MAP = os.path.join(CONFIG_USER_FILES, 'kicad_map.yaml')

# Inventree
CONFIG_CATEGORIES = os.path.join(CONFIG_USER_FILES, 'categories.yaml')
CONFIG_PARAMETERS = os.path.join(CONFIG_USER_FILES, 'parameters.yaml')
CONFIG_PARAMETERS_FILTERS = os.path.join(
    CONFIG_USER_FILES, 'parameters_filters.yaml')

# INTERNAL PART NUMBERS
CONFIG_IPN = config_interface.load_file(os.path.join(CONFIG_USER_FILES, 'internal_part_number.yaml'))
IPN_UNIQUE_ID_LENGTH = CONFIG_IPN.get('IPN_UNIQUE_ID_LENGTH', 6)
IPN_USE_FIXED_PREFIX = CONFIG_IPN.get('IPN_USE_FIXED_PREFIX', False)
if IPN_USE_FIXED_PREFIX:
    IPN_PREFIX = CONFIG_IPN.get('IPN_PREFIX', '')
IPN_USE_VARIANT_SUFFIX = CONFIG_IPN.get('IPN_USE_VARIANT_SUFFIX', True)
if IPN_USE_VARIANT_SUFFIX:
    IPN_VARIANT_SUFFIX = CONFIG_IPN.get('IPN_VARIANT_SUFFIX', '00')

# GENERAL SETTINGS
CONFIG_GENERAL = config_interface.load_file(os.path.join(CONFIG_USER_FILES, 'general.yaml'))
AUTOMATIC_SUBCATEGORY_CREATE = CONFIG_GENERAL.get('AUTOMATIC_SUBCATEGORY_CREATE', False)
AUTOMATIC_BROWSER_OPEN = CONFIG_GENERAL.get('AUTOMATIC_BROWSER_OPEN', False)
DEFAULT_SUPPLIER = CONFIG_GENERAL.get('DEFAULT_SUPPLIER', 'Digi-Key')
# Load enable flags
try:
    ENABLE_KICAD = CONFIG_GENERAL.get('ENABLE_KICAD', False)
    ENABLE_INVENTREE = CONFIG_GENERAL.get('ENABLE_INVENTREE', False)
    ENABLE_ALTERNATE = CONFIG_GENERAL.get('ENABLE_ALTERNATE', False)
except TypeError:
    pass

# Supported suppliers APIs
SUPPORTED_SUPPLIERS_API = [
    'Digi-Key',
    'Mouser',
    'Farnell',
    'Newark',
    'Element14',
    'LCSC',
]

# Generic API user configuration
CONFIG_SUPPLIER_PARAMETERS = os.path.join(CONFIG_USER_FILES, 'supplier_parameters.yaml')
CONFIG_SEARCH_API = config_interface.load_file(os.path.join(CONFIG_USER_FILES, 'search_api.yaml'))

# Digi-Key user configuration
if 'Digi-Key' in SUPPORTED_SUPPLIERS_API:
    CONFIG_DIGIKEY = config_interface.load_file(os.path.join(CONFIG_USER_FILES, 'digikey_config.yaml'))
    CONFIG_DIGIKEY_API = os.path.join(CONFIG_USER_FILES, 'digikey_api.yaml')
    CONFIG_DIGIKEY_CATEGORIES = os.path.join(CONFIG_USER_FILES, 'digikey_categories.yaml')
    # CONFIG_DIGIKEY_PARAMETERS = os.path.join(CONFIG_USER_FILES, 'digikey_parameters.yaml')

# Mouser user configuration
if 'Mouser' in SUPPORTED_SUPPLIERS_API:
    CONFIG_MOUSER = config_interface.load_file(os.path.join(CONFIG_USER_FILES, 'mouser_config.yaml'))
    CONFIG_MOUSER_API = os.path.join(CONFIG_USER_FILES, 'mouser_api.yaml')

# Element14 user configuration (includes Farnell, Newark and Element14)
if 'Element14' in SUPPORTED_SUPPLIERS_API:
    CONFIG_ELEMENT14 = config_interface.load_file(os.path.join(CONFIG_USER_FILES, 'element14_config.yaml'))
    CONFIG_ELEMENT14_API = os.path.join(CONFIG_USER_FILES, 'element14_api.yaml')

# LCSC user configuration
if 'LCSC' in SUPPORTED_SUPPLIERS_API:
    CONFIG_LCSC = config_interface.load_file(os.path.join(CONFIG_USER_FILES, 'lcsc_config.yaml'))
    CONFIG_LCSC_API = os.path.join(CONFIG_USER_FILES, 'lcsc_api.yaml')

# Automatic category match confidence level (from 0 to 100)
CATEGORY_MATCH_RATIO_LIMIT = CONFIG_SEARCH_API.get('CATEGORY_MATCH_RATIO_LIMIT', 100)
# Search results caching (stored in files)
CACHE_ENABLED = CONFIG_SEARCH_API.get('CACHE_ENABLED', True)
# Cache validity in days
CACHE_VALID_DAYS = CONFIG_SEARCH_API.get('CACHE_VALID_DAYS', 7)


# Caching settings
def load_cache_settings():
    global search_results
    global search_images
    global CACHE_ENABLED
    global DIGIKEY_STORAGE_PATH
    
    USER_SETTINGS = config_interface.load_user_paths(home_dir=HOME_DIR)

    if CACHE_ENABLED:
        search_results = {
            'directory': os.path.join(USER_SETTINGS['USER_CACHE'], 'search', ''),
            'extension': '.yaml',
        }
        # Create folder if it does not exists
        if not os.path.exists(search_results['directory']):
            os.makedirs(search_results['directory'])

    # Part images
    search_images = os.path.join(USER_SETTINGS['USER_CACHE'], 'images', '')
    # Create folder if it does not exists
    if not os.path.exists(search_images):
        os.makedirs(search_images)

    # API token storage path
    DIGIKEY_STORAGE_PATH = os.path.join(USER_SETTINGS['USER_CACHE'], '')


# Load cache settings
load_cache_settings()

# KICAD
# User Settings
KICAD_SYMBOLS_PATH = ''
KICAD_TEMPLATES_PATH = ''
KICAD_FOOTPRINTS_PATH = ''


def load_kicad_settings():
    global CONFIG_KICAD
    global KICAD_SYMBOLS_PATH
    global KICAD_TEMPLATES_PATH
    global KICAD_FOOTPRINTS_PATH

    kicad_user_settings = config_interface.load_file(CONFIG_KICAD, silent=False)
    if kicad_user_settings:
        KICAD_SYMBOLS_PATH = kicad_user_settings.get('KICAD_SYMBOLS_PATH', None)
        KICAD_TEMPLATES_PATH = kicad_user_settings.get('KICAD_TEMPLATES_PATH', None)
        KICAD_FOOTPRINTS_PATH = kicad_user_settings.get('KICAD_FOOTPRINTS_PATH', None)


# Load kicad settings
load_kicad_settings()


def set_default_supplier(value: str, save=False):
    global DEFAULT_SUPPLIER
    DEFAULT_SUPPLIER = value
    if save:
        user_settings = config_interface.load_file(os.path.join(CONFIG_USER_FILES, 'general.yaml'))
        user_settings['DEFAULT_SUPPLIER'] = value
        config_interface.dump_file(user_settings, os.path.join(CONFIG_USER_FILES, 'general.yaml'))
    return


# Library Paths
if not ENABLE_TEST:
    symbol_libraries_paths = config_interface.load_libraries_paths(
        CONFIG_KICAD_CATEGORY_MAP, KICAD_SYMBOLS_PATH)
# cprint(symbol_libraries_paths)

# Template Paths
symbol_templates_paths = config_interface.load_templates_paths(
    CONFIG_KICAD_CATEGORY_MAP, KICAD_TEMPLATES_PATH)
# cprint(symbol_templates_paths)

# Footprint Libraries
footprint_libraries_paths = config_interface.load_footprint_paths(
    CONFIG_KICAD_CATEGORY_MAP, KICAD_FOOTPRINTS_PATH)
# cprint(footprint_libraries_paths)
footprint_name_default = 'TBD'

AUTO_GENERATE_LIB = True
symbol_template_lib = os.path.join(
    PROJECT_DIR, 'kicad', 'templates', 'library_template.kicad_sym')


# INVENTREE
class Environment(Enum):
    '''
    Server/Remote Development: DEVELOPMENT
    Server/Remote Production: PRODUCTION
    '''
    DEVELOPMENT = 0
    PRODUCTION = 1


# Pick environment
environment = CONFIG_GENERAL.get('INVENTREE_ENV', None)
environment = os.environ.get('INVENTREE_ENV', environment)

try:
    environment = int(environment)
except TypeError:
    environment = 0

# Load correct user file
if environment == Environment.PRODUCTION.value:
    CONFIG_INVENTREE = os.path.join(CONFIG_USER_FILES, 'inventree_prod.yaml')
else:
    CONFIG_INVENTREE = os.path.join(CONFIG_USER_FILES, 'inventree_dev.yaml')

# Load user settings
inventree_settings = config_interface.load_inventree_user_settings(CONFIG_INVENTREE)


# Server settings
def load_inventree_settings():
    global SERVER_ADDRESS
    global USERNAME
    global PASSWORD
    global PART_URL_ROOT

    inventree_settings = config_interface.load_inventree_user_settings(CONFIG_INVENTREE)

    SERVER_ADDRESS = inventree_settings.get('SERVER_ADDRESS', None)
    USERNAME = inventree_settings.get('USERNAME', None)
    PASSWORD = inventree_settings.get('PASSWORD', None)
    # Part URL
    if SERVER_ADDRESS:
        # If missing, append slash to root URL
        root_url = SERVER_ADDRESS
        if not SERVER_ADDRESS.endswith('/'):
            root_url = root_url + '/'
        # Set part URL
        PART_URL_ROOT = root_url + 'part/'


# Default revision
INVENTREE_DEFAULT_REV = inventree_settings.get('INVENTREE_DEFAULT_REV', 'A')

# InvenTree part dictionary template
inventree_part_template = {
    'category': [None, None],
    'name': None,
    'description': None,
    'IPN': None,
    'revision': None,
    'keywords': None,
    'image': None,
    'inventree_url': None,
    'manufacturer': {},
    'datasheet': None,
    'supplier': {},
    'supplier_link': None,
    'parameters': {},
}


# Enable flags
def set_enable_flags(values: list):
    global CONFIG_GENERAL
    global ENABLE_KICAD
    global ENABLE_INVENTREE
    global ENABLE_ALTERNATE

    ENABLE_KICAD = values[0]
    ENABLE_INVENTREE = values[1]
    ENABLE_ALTERNATE = values[2]

    # Save user settings
    user_settings = CONFIG_GENERAL
    user_settings['ENABLE_KICAD'] = ENABLE_KICAD
    user_settings['ENABLE_INVENTREE'] = ENABLE_INVENTREE
    user_settings['ENABLE_ALTERNATE'] = ENABLE_ALTERNATE
    config_interface.dump_file(user_settings, os.path.join(CONFIG_USER_FILES, 'general.yaml'))

    return



name = 'wrapt_timeout_decorator'
title = 'there are many timeout decorators out there - that one focuses on correctness when using with Classes, methods, class methods, static methods and so on'
version = '1.4.0'
url = 'https://github.com/bitranox/wrapt_timeout_decorator'
author = 'Robert Nowotny'
author_email = 'bitranox@gmail.com'
shell_command = 'wrapt_timeout_decorator'


def print_version() -> None:
    print('version: 1.4.0')


def print_info() -> None:
    print("""information for "wrapt_timeout_decorator":

          there are many timeout decorators out there - that one focuses on correctness when using with Classes, methods, class methods, static methods and so on

          Version      : 1.4.0
          url          : https://github.com/bitranox/wrapt_timeout_decorator
          author       : Robert Nowotny
          author_email : bitranox@gmail.com""")



import os

from ..config import settings, config_interface
from mouser.api import MouserPartSearchRequest

SEARCH_HEADERS = [
    'Description',
    'productCode',
    'MouserPartNumber',
    'Manufacturer',
    'ManufacturerPartNumber',
    'DataSheetUrl',
    'ProductDetailUrl',
    'ImagePath',
]
PARAMETERS_MAP = [
    'ProductAttributes',
    'AttributeName',
    'AttributeValue',
]


def get_default_search_keys():
    return [
        'Description',
        'Description',
        'revision',
        'keywords',
        'MouserPartNumber',
        'Manufacturer',
        'ManufacturerPartNumber',
        'ProductDetailUrl',
        'DataSheetUrl',
        'ImagePath',
    ]


def setup_environment(force=False):
    ''' Setup environmental variables '''

    api_key = os.environ.get('MOUSER_PART_API_KEY', None)
    if not api_key or force:
        mouser_api_settings = config_interface.load_file(settings.CONFIG_MOUSER_API)
        os.environ['MOUSER_PART_API_KEY'] = mouser_api_settings['MOUSER_PART_API_KEY']


def find_categories(part_details: str):
    ''' Find categories '''

    try:
        return part_details['Category'], None
    except:
        return None, None


def fetch_part_info(part_number: str) -> dict:
    ''' Fetch part data from API '''

    from ..wrapt_timeout_decorator import timeout

    setup_environment()
    part_info = {}

    @timeout(dec_timeout=20)
    def search_timeout():
        request = MouserPartSearchRequest('partnumber')
        request.part_search(part_number)
        return request.get_clean_response()

    # Query part number
    try:
        part = search_timeout()
    except:
        part = None

    if not part:
        return part_info

    # Check for empty response
    empty = True
    for key, value in part.items():
        if value:
            empty = False
            break
    if empty:
        return part_info

    category, subcategory = find_categories(part)
    try:
        part_info['category'] = category
        part_info['subcategory'] = subcategory
    except:
        part_info['category'] = ''
        part_info['subcategory'] = ''

    headers = SEARCH_HEADERS

    for key in part:
        if key in headers:
            part_info[key] = part[key]

    # Parameters
    part_info['parameters'] = {}
    [parameter_key, name_key, value_key] = PARAMETERS_MAP

    for parameter in range(len(part[parameter_key])):
        parameter_name = part[parameter_key][parameter][name_key]
        parameter_value = part[parameter_key][parameter][value_key]
        # Append to parameters dictionary
        part_info['parameters'][parameter_name] = parameter_value

    return part_info


def test_api() -> bool:
    ''' Test method for API '''

    test_success = True
    expected = {
        'Description': 'MOSFET P-channel 1.25W',
        'MouserPartNumber': '621-DMP2066LSN-7',
        'Manufacturer': 'Diodes Incorporated',
        'ManufacturerPartNumber': 'DMP2066LSN-7',
    }

    test_part = fetch_part_info('DMP2066LSN-7')
        
    if not test_part:
        # Unsucessful search
        test_success = False
    else:
        # Check content of response
        for key, value in expected.items():
            if test_part[key] != value:
                print(f'"{test_part[key]}" <> "{value}"')
                test_success = False
                break

    return test_success



# STDLIB
import sys
from typing import Any

# EXT
import multiprocess     # type: ignore

# OWN
from .wrap_helper import WrapHelper, raise_exception


class Timeout(object):
    """Wrap a function and add a timeout (limit) attribute to it.
    Instances of this class are automatically generated by the add_timeout
    function defined above. Wrapping a function allows asynchronous calls
    to be made and termination of execution after a timeout has passed.
    """

    def __init__(self, wrap_helper: WrapHelper) -> None:
        """Initialize instance in preparation for being called."""
        self.wrap_helper = wrap_helper
        self.__name__ = self.wrap_helper.wrapped.__name__
        self.__doc__ = self.wrap_helper.wrapped.__doc__
        self.__process = None       # type: multiprocess.Process
        self.__parent_conn = None   # type: multiprocess.Pipe

    def __call__(self) -> Any:
        """Execute the embedded function object asynchronously.
        The function given to the constructor is transparently called and
        requires that "ready" be intermittently polled. If and when it is
        True, the "value" property may then be checked for returned data.
        """
        self.__parent_conn, self.wrap_helper.child_conn = multiprocess.Pipe(duplex=False)
        self.__process = multiprocess.Process(target=_target, args=[self.wrap_helper])
        # daemonic process must not have subprocess - we need that for nested decorators
        self.__process.daemon = False
        self.__process.start()
        if not self.wrap_helper.dec_hard_timeout:
            self.wait_until_process_started()
        if self.__parent_conn.poll(self.wrap_helper.dec_timeout_float):
            return self.value
        else:
            self.cancel()

    def cancel(self) -> None:
        """Terminate any possible execution of the embedded function."""
        if self.__process.is_alive():   # pragma: no cover      # we can not produce that state - its just a security measure
            self.__process.terminate()
        self.__process.join(timeout=1.0)
        self.__parent_conn.close()
        raise_exception(self.wrap_helper.timeout_exception, self.wrap_helper.exception_message)

    def wait_until_process_started(self) -> None:
        self.__parent_conn.recv()

    @property
    def value(self) -> Any:
        exception_occured, result = self.__parent_conn.recv()
        # when self.__parent_conn.recv() exits, maybe __process is still alive,
        # then it might zombie the process. so join it explicitly
        self.__process.join(timeout=1.0)
        self.__parent_conn.close()

        if exception_occured:
            raise result
        else:
            return result


def _target(wrap_helper: WrapHelper) -> None:
    """Run a function with arguments and return output via a pipe.
    This is a helper function for the Process created in Timeout. It runs
    the function with positional arguments and keyword arguments and then
    returns the function's output by way of a queue. If an exception gets
    raised, it is returned to Timeout to be raised by the value property.
    """
    # noinspection PyBroadException
    try:
        if not wrap_helper.dec_hard_timeout:
            wrap_helper.child_conn.send('started')
        exception_occured = False
        wrap_helper.child_conn.send((exception_occured, wrap_helper.wrapped(*wrap_helper.args, **wrap_helper.kwargs)))
    except Exception:
        exception_occured = True
        wrap_helper.child_conn.send((exception_occured, sys.exc_info()[1]))
    finally:
        wrap_helper.child_conn.close()



# STDLIB
import logging
import platform
import signal
import sys
import threading
from types import FrameType
from typing import Any, Callable, Dict, List, Type, Union

# EXT
import dill             # type: ignore
import multiprocess     # type: ignore

logging.basicConfig()
logger = logging.getLogger('pickle_analyzer')


class WrapHelper(object):
    def __init__(self,
                 dec_timeout: Union[None, float, str],
                 use_signals: bool,
                 timeout_exception: Type[BaseException],
                 exception_message: str,
                 dec_allow_eval: bool,
                 dec_hard_timeout: bool,
                 wrapped: Callable[..., Any],
                 instance: object,
                 args: Any,
                 kwargs: Any) -> None:
        self.dec_timeout = dec_timeout
        self.use_signals = use_signals
        self.timeout_exception = timeout_exception
        self.exception_message = exception_message
        self.dec_allow_eval = dec_allow_eval
        self.dec_hard_timeout = dec_hard_timeout
        self.wrapped = wrapped
        self.instance = instance
        self.args = args
        self.kwargs = kwargs

        self.dec_timeout_float = 0.0                # type: float
        self.old_alarm_handler = None               # type: Union[Callable[[signal.Signals, FrameType], None], int, signal.Handlers, None]
        self.child_conn = None                      # type: multiprocess.Pipe

        self.pop_kwargs()
        self.set_signals_to_false_if_not_possible()
        self.eval_if_required()
        self.convert_timeout_given_to_float()
        self.format_exception_message()

    def convert_timeout_given_to_float(self) -> None:
        if self.dec_timeout is None:
            self.dec_timeout_float = 0.0
        else:
            try:
                self.dec_timeout_float = float(self.dec_timeout)
            except ValueError:
                raise ValueError('the given or evaluated value for the timeout can not be converted to float : "{dec_timeout}"'
                                 .format(dec_timeout=self.dec_timeout))

    def pop_kwargs(self) -> None:
        self.dec_allow_eval = self.kwargs.pop('dec_allow_eval', self.dec_allow_eval)
        self.dec_timeout = self.kwargs.pop('dec_timeout', self.dec_timeout)
        self.use_signals = self.kwargs.pop('use_signals', self.use_signals)
        self.dec_hard_timeout = self.kwargs.pop('dec_hard_timeout', self.dec_hard_timeout)

    @property
    def should_eval(self) -> bool:
        if self.dec_allow_eval and isinstance(self.dec_timeout, str):
            return True
        else:
            return False

    def format_exception_message(self) -> None:
        function_name = self.wrapped.__name__ or '(unknown name)'
        if not self.exception_message:
            self.exception_message = 'Function {function_name} timed out after {dec_timeout_float} seconds'
        self.exception_message = self.exception_message.format(function_name=function_name, dec_timeout_float=self.dec_timeout_float)

    def new_alarm_handler(self, signum: signal.Signals, frame: FrameType) -> None:
        raise_exception(self.timeout_exception, self.exception_message)

    def save_old_and_set_new_alarm_handler(self) -> None:
        self.old_alarm_handler = signal.signal(signal.SIGALRM, self.new_alarm_handler)
        signal.setitimer(signal.ITIMER_REAL, self.dec_timeout_float)

    def restore_old_alarm_handler(self) -> None:
        signal.setitimer(signal.ITIMER_REAL, 0)
        signal.signal(signal.SIGALRM, self.old_alarm_handler)

    def set_signals_to_false_if_not_possible(self) -> None:
        if is_system_windows() or not is_in_main_thread():
            self.use_signals = False

    def eval_if_required(self) -> None:
        wrapped = self.wrapped
        instance = self.instance
        args = self.args
        kwargs = self.kwargs

        if self.should_eval:
            self.dec_timeout = eval(str(self.dec_timeout))


def detect_unpickable_objects_and_reraise(object_to_pickle: Any) -> None:
    # sometimes the detection detects unpickable objects but actually
    # they can be pickled - so we just try to start the thread and report
    # the unpickable objects if that fails
    dict_result = detect_unpickable_objects(object_to_pickle, dill_trace=False, log_warning=False)
    s_err = 'can not pickle {on}, bad items: {bi}, bad objects: {bo}, bad types {bt}'.format(on=dict_result['object_name'],
                                                                                             bi=dict_result['bad_items'],
                                                                                             bo=dict_result['bad_objects'],
                                                                                             bt=dict_result['bad_types'])
    raise dill.PicklingError(s_err)


def detect_unpickable_objects(object_to_pickle: Any, dill_trace: bool = True, log_warning: bool = True) -> Dict[str, Union[str, List[Any]]]:
    if log_warning:
        logger.warning('always remember that the "object_to_pickle" should not be defined within the main context')
    dict_result = dict()                # type: Dict[str, Union[str, List[Any]]]
    dict_result['object_name'] = ''
    dict_result['bad_items'] = list()
    dict_result['bad_objects'] = list()
    dict_result['bad_types'] = list()
    safe_status_of_dill_trace = dill.detect.trace
    # noinspection PyBroadException
    try:
        if dill_trace:
            dill.detect.trace = True
        pickled_object = dill.dumps(object_to_pickle)
        dill.loads(pickled_object)
    except Exception:
        dict_result['object_name'] = get_object_name(object_to_pickle)
        dict_result['bad_objects'] = get_bad_pickling_objects(object_to_pickle)
        dict_result['bad_types'] = get_bad_pickling_types(object_to_pickle)
    finally:
        dill.detect.trace = safe_status_of_dill_trace
        return dict_result


def get_object_name(object_to_pickle: object) -> str:
    object_name = 'object'
    if hasattr(object_to_pickle, '__name__'):
        if object_to_pickle.__name__:                   # type: ignore
            object_name = object_to_pickle.__name__     # type: ignore
    return object_name


def get_bad_pickling_types(object_to_pickle: object) -> List[Any]:
    bad_types = list()                                  # type: List[Any]
    # noinspection PyBroadException
    try:
        bad_types = dill.detect.badtypes(object_to_pickle)
    except Exception:
        bad_types = [sys.exc_info()[1]]
    finally:
        return bad_types


def get_bad_pickling_objects(object_to_pickle: Any) -> Any:
    bad_objects = list()                                # type: List[object]
    # noinspection PyBroadException
    try:
        bad_objects = dill.detect.badobjects(object_to_pickle)
    except Exception:
        bad_objects = [sys.exc_info()[1]]
    finally:
        return bad_objects


def raise_exception(exception: Type[BaseException], exception_message: str) -> None:
    """ This function checks if a exception message is given.
    If there is no exception message, the default behaviour is maintained.
    If there is an exception message, the message is passed to the exception.
    """
    if not exception:
        exception = TimeoutError
    raise exception(exception_message)


def is_in_main_thread() -> bool:
    if threading.current_thread() == threading.main_thread():
        return True
    else:
        return False


def is_system_windows() -> bool:
    if platform.system().lower().startswith('win'):
        return True
    else:
        return False



"""
Timeout decorator.
    :copyright: (c) 2017 by Robert Nowotny
    :license: MIT, see LICENSE for more details.
"""

# STDLIB
from typing import Any, Callable, Type, Union

# EXT
from dill import PicklingError          # type: ignore
import wrapt                            # type: ignore

# OWN
from .wrap_helper import WrapHelper, detect_unpickable_objects_and_reraise
from .wrap_function_multiprocess import Timeout


def timeout(dec_timeout: Union[None, float, str] = None,
            use_signals: bool = True,
            timeout_exception: Type[BaseException] = TimeoutError,
            exception_message: str = '',
            dec_allow_eval: bool = False,
            dec_hard_timeout: bool = False) -> Any:

    """Add a timeout parameter to a function and return it.

    ToDo :   not clear how to type a decorator factory,
             tried:   ->  Callable[..., Any]
                ...
             return cast(Callable[..., Any], wrapped)
             without success - so we stuck with any at the moment
             ** see example on bottom of that file for correct annotation of a generic decorator

    Windows remark : dont use the decorator on classes in the main.py because of Windows multiprocessing limitations
                     read the README

    Usage:

    @timeout(3)
    def foo():
        pass

    Overriding the timeout:

    foo(dec_timeout=5)

    Usage without decorating a function :

    def test_method(a,b,c):
        pass

    timeout(3)(test_method)(1,2,c=3)

    Usage with eval (beware, security hazard, no user input values here):
        read : https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html before usage !

    def class ClassTest4(object):
        def __init__(self,x):
            self.x=x

        @timeout('instance.x', dec_allow_eval=True)
        def test_method(self):
            print('swallow')

        @timeout(1)
        def foo3(self):
            print('parrot')

    # or override via kwarg :
    my_foo = ClassTest4(3)
    my_foo.test_method(dec_timeout='instance.x * 2.5 +1')
    my_foo.foo3(dec_timeout='instance.x * 2.5 +1', dec_allow_eval=True)

    :param dec_timeout: *       optional time limit in seconds or fractions of a second. If None is passed,
                                no seconds is applied. This adds some flexibility to the usage: you can disable timing
                                out depending on the settings. dec_timeout will always be overridden by a
                                kwarg passed to the wrapped function, class or class method.
    :param use_signals:         flag indicating whether signals should be used or the multiprocessing
    :param timeout_exception:   the Exception to be raised when timeout occurs, default = TimeoutException
    :param exception_message:   the Message for the Exception. Default: 'Function {f} timed out after {s} seconds.
    :param dec_allow_eval: *    allows a string in parameter dec_timeout what will be evaluated. Beware this can
                                be a security issue. This is very powerful, but is also very dangerous if you
                                accept strings to evaluate from untrusted input.
                                read: https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html

                                If enabled, the parameter of the function dec_timeout, or the parameter passed
                                by kwarg dec_timeout will be evaluated if its type is string. You can access :
                                wrapped (the function object and all their exposed objects)
                                instance    Example: 'instance.x' - see example above or doku
                                args        Example: 'args[0]' - the timeout is the first argument in args
                                kwargs      Example: 'kwargs["max_time"] * 2'

    :param dec_hard_timeout:    only considered when use_signals = True (Windows)
                                if dec_hard_timeout = True, the decorator will timeout after dec_timeout after the
                                decorated function is called by the main program.
                                If You set up a small timeout value like 0.1 seconds, in windows that function might
                                actually never run - because setting up the process will already take longer
                                then 0.1 seconds - that means the decorated function will ALWAYS time out (and never run).

                                if dec_hard_timeout = False, the decorator will timeout after the process is allowed to
                                run for dec_timeout seconds, that means the time to set up the new process is not considered.
                                If You set up a small timeout value like 0.1 seconds, in windows that function might now
                                take something like 0.6 seconds to timeout - 0.5 seconds to set up the process, and
                                allowing the function in the process to run for 0.1 seconds.
                                Since You can not know how long the spawn() will take under Windows, this is the default setting.

    * all parameters starting with dec_ can be overridden via kwargs passed to the wrapped function.

    :raises:                    TimeoutError if time limit is reached
    :returns:                   the Result of the wrapped function

    It is illegal to pass anything other than a function as the first parameter.
    The function is wrapped and returned to the caller.
    """

    @wrapt.decorator    # type: ignore
    def wrapper(wrapped: Callable[..., Any], instance: object, args: Any, kwargs: Any) -> Any:
        wrap_helper = WrapHelper(dec_timeout, use_signals, timeout_exception, exception_message, dec_allow_eval,
                                 dec_hard_timeout, wrapped, instance, args, kwargs)
        if not wrap_helper.dec_timeout_float:
            return wrapped(*wrap_helper.args, **wrap_helper.kwargs)
        else:
            return wrapped_with_timeout(wrap_helper)
    return wrapper


def wrapped_with_timeout(wrap_helper: WrapHelper) -> Any:
    if wrap_helper.use_signals:
        return wrapped_with_timeout_signals(wrap_helper)
    else:
        return wrapped_with_timeout_process(wrap_helper)


def wrapped_with_timeout_signals(wrap_helper: WrapHelper) -> Any:
    try:
        wrap_helper.save_old_and_set_new_alarm_handler()
        return wrap_helper.wrapped(*wrap_helper.args, **wrap_helper.kwargs)
    finally:
        wrap_helper.restore_old_alarm_handler()


def wrapped_with_timeout_process(wrap_helper: WrapHelper) -> Any:
    try:
        timeout_wrapper = Timeout(wrap_helper)
        return timeout_wrapper()
    except PicklingError:
        detect_unpickable_objects_and_reraise(wrap_helper.wrapped)


"""

# Example for generic decorator with does not destroy the signature of the wrapped function for mypy 

from typing import Any, Callable, TypeVar, cast

F = TypeVar('F', bound=Callable[..., Any])


def check_for_kwargs(f: F) -> F:
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if kwargs:
            keys = ', '.join([key for key in kwargs.keys()])
            raise TypeError("{fn}() got some positional-only arguments passed as keyword arguments: '{keys}'".format(fn=f.__name__, keys=keys))
        return f(*args, **kwargs)
    return cast(F, wrapper)

"""


import time

# PySimpleGUI
import PySimpleGUI as sg

CREATE_PART_PROGRESS: int
MAX_PROGRESS = 100
DEFAULT_PROGRESS = 5


def create_progress_bar_window(font=None, location=None) -> bool:
    ''' Create window for part creation progress '''
    global CREATE_PART_PROGRESS, MAX_PROGRESS
    global progress_window

    progress_layout = [
        [sg.Text('Creating part...')],
        [sg.ProgressBar(MAX_PROGRESS, orientation='h', size=(20, 30), key='progressbar')],
        [sg.Cancel()]
    ]
    progress_window = sg.Window('Part Creation Progress', progress_layout, font=font, location=location)
    # progress_bar = progress_window['progressbar']

    event, values = progress_window.read(timeout=10)

    # Reset progress
    CREATE_PART_PROGRESS = 0
    update_progress_bar_window()

    if progress_window:
        return True

    return False


def close_progress_bar_window():
    ''' Close progress bar window after part creation '''
    global progress_window

    try:
        if progress_window:
            progress_window.close()
    except NameError:
        return


def progress_increment():
    ''' Increment progress '''
    global CREATE_PART_PROGRESS, MAX_PROGRESS

    if CREATE_PART_PROGRESS + 1 < MAX_PROGRESS:
        CREATE_PART_PROGRESS += 1
    else:
        CREATE_PART_PROGRESS = MAX_PROGRESS


def update_progress_bar_window(increment=0) -> bool:
    ''' Update progress bar during part creation '''
    global CREATE_PART_PROGRESS, MAX_PROGRESS, DEFAULT_PROGRESS
    global progress_window

    on_going_progress = True

    try:
        if not progress_window:
            return False
    except NameError:
        return False

    if increment:
        inc = increment
    else:
        # Default
        inc = DEFAULT_PROGRESS

    progress_bar = progress_window['progressbar']

    event, values = progress_window.read(timeout=10)

    if event in ['Cancel', sg.WIN_CLOSED]:
        on_going_progress = False
        close_progress_bar_window()
    else:
        # Smooth effect
        for i in range(inc):
            progress_increment()
            progress_bar.update(CREATE_PART_PROGRESS, MAX_PROGRESS)
            if inc < MAX_PROGRESS:
                time.sleep(0.02)
            else:
                time.sleep(0.005)

    return on_going_progress



""" DISCLAIMER
Sexpr classes and methods were directly imported from: https://gitlab.com/kicad/libraries/kicad-library-utils
Last import date: Jan 3rd, 2023
"""

"""
code extracted from: http://rosettacode.org/wiki/S-Expressions
"""

import re
from typing import Any, Optional

dbg: bool = False

term_regex = r"""(?mx)
    \s*(?:
        (\()|
        (\))|
        ([+-]?\d+\.\d+(?=[\ \)]))|
        (\-?\d+(?=[\ \)]))|
        "((?:[^"]|(?<=\\)")*)"|
        ([^(^)\s]+)
       )"""


class SexprError(ValueError):
    pass


def parse_sexp(sexp: str) -> Any:
    re_iter = re.finditer(term_regex, sexp)
    rv = list(_parse_sexp_internal(re_iter))

    for leftover in re_iter:
        lparen, rparen, *rest = leftover.groups()
        if lparen or any(rest):
            raise SexprError(f'Leftover garbage after end of expression at position {leftover.start()}')  # noqa: E501

        elif rparen:
            raise SexprError(f'Unbalanced closing parenthesis at position {leftover.start()}')

    if len(rv) == 0:
        raise SexprError('No or empty expression')

    if len(rv) > 1:
        raise SexprError('Missing initial opening parenthesis')

    return rv[0]


def _parse_sexp_internal(re_iter) -> Any:
    for match in re_iter:
        lparen, rparen, float_num, integer_num, quoted_str, bare_str = match.groups()

        if lparen:
            yield list(_parse_sexp_internal(re_iter))
        elif rparen:
            break
        elif bare_str is not None:
            yield bare_str
        elif quoted_str is not None:
            yield quoted_str.replace('\\"', '"')
        elif float_num:
            yield float(float_num)
        elif integer_num:
            yield int(integer_num)


# Form a valid sexpr (single line)
def SexprItem(val: Any, key: Optional[str] = None) -> str:
    if key:
        fmt = "(" + key + " {val})"
    else:
        fmt = "{val}"

    if val is None or isinstance(val, str) and len(val) == 0:
        val = '""'
    elif isinstance(val, (list, tuple)):
        val = " ".join([SexprItem(v) for v in val])
    elif isinstance(val, dict):
        values = []
        for key in val.keys():
            values.append(SexprItem(val[key], key))
        val = " ".join(values)
    elif isinstance(val, float):
        val = str(round(val, 10)).rstrip("0").rstrip(".")
    elif isinstance(val, int):
        val = str(val)
    elif isinstance(val, str) and re.search(r"[\s()\"]", val):
        val = '"%s"' % repr(val)[1:-1].replace('"', r"\"")

    return fmt.format(val=val)


class SexprBuilder:
    def __init__(self, key):
        self.indent: int = 0
        self.output: str = ""
        self.items = []
        if key is not None:
            self.startGroup(key, newline=False)

    def _indent(self) -> None:
        self.output += " " * 2 * self.indent

    def _newline(self) -> None:
        self.output += "\n"

    def _addItems(self) -> None:
        self.output += " ".join(str(i) for i in self.items)
        self.items = []

    def startGroup(
        self, key: Optional[Any] = None, newline: bool = True, indent: bool = False
    ) -> None:
        self._addItems()
        if newline and indent:
            self.indent += 1
        if newline:
            self._newline()
            self._indent()
        self.output += "("
        if key:
            self.output += str(key) + " "

    def endGroup(self, newline: bool = True) -> None:
        self._addItems()
        if newline:
            self._newline()
            if self.indent > 0:
                self.indent -= 1
            self._indent()
        self.output += ")"

    def addOptItem(self, key, item, newline: bool = True, indent: bool = False):
        if item in [None, 0, False]:
            return

        self.addItems({key: item}, newline=newline, indent=indent)

    def addItem(self, item, newline: bool = True, indent: bool = False):
        self._addItems()
        if newline and indent:
            self.indent += 1
        if newline:
            self.newLine()
        self.items.append(SexprItem(item))

    # Add a (preformatted) item
    def addItems(self, items, newline: bool = True, indent: bool = False):
        self._addItems()
        if indent:
            self.indent += 1
        if newline:
            self.newLine()
        if isinstance(items, (list, tuple)):
            for item in items:
                self.items.append(SexprItem(item))
        else:
            self.items.append(SexprItem(items))

    def newLine(self, indent: bool = False) -> None:
        self._addItems()
        self._newline()
        if indent:
            self.indent += 1
        self._indent()

    def unIndent(self) -> None:
        if self.indent > 0:
            self.indent -= 1


def build_sexp(exp, indent='  ') -> str:
    # Special case for multi-values
    if isinstance(exp, list):
        joined = '('
        for i, elem in enumerate(exp):
            if 1 <= i <= 5 and len(joined) < 120 and not isinstance(elem, list):
                joined += ' '
            elif i >= 1:
                joined += '\n' + indent
            joined += build_sexp(elem, indent=f'{indent}  ')
        return joined + ')'

    if exp == '':
        return '""'

    return str(exp)


def format_sexp(sexp: str, indentation_size: int = 2, max_nesting: int = 2) -> str:
    out = ''
    n = 0
    for match in re.finditer(term_regex, sexp):
        indentation = ""
        lparen, rparen, float_num, integer_num, quoted_str, bare_str = match.groups()
        if lparen:
            if out:
                if n <= max_nesting:
                    if out[-1] == ' ':
                        out = out[:-1]
                    indentation = '\n' + (' ' * indentation_size * n)
                else:
                    if out[-1] == ')':
                        out += ' '
            n += 1
            out += indentation + '('
        elif rparen:
            if out and out[-1] == ' ':
                out = out[:-1]
            n -= 1
            out += indentation + ')'
        elif float_num:
            out += indentation + float_num + ' '
        elif integer_num:
            out += indentation + integer_num + ' '
        elif quoted_str is not None:
            out += f'{indentation}"{quoted_str}" '
        elif bare_str is not None:
            out += indentation + bare_str + ' '

    out += '\n'
    return out


if __name__ == "__main__":
    sexp = """ ( ( data "quoted data" 123 4.5)
         (data "with \\"escaped quotes\\"")
         (data (123 (4.5) "(more" "data)")))"""

    print("Input S-expression:")
    print(sexp)
    parsed = parse_sexp(sexp)
    print("\nParsed to Python:", parsed)

    print("\nThen back to: '%s'" % build_sexp(parsed))
    print("\nThen back to: '%s'" % format_sexp(build_sexp(parsed)))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kintree',
 'kintree.common',
 'kintree.config',
 'kintree.database',
 'kintree.kicad',
 'kintree.kicad.lib_utils',
 'kintree.search',
 'kintree.wrapt_timeout_decorator']

package_data = \
{'': ['*'],
 'kintree.config': ['digikey/*',
                    'element14/*',
                    'inventree/*',
                    'kicad/*',
                    'lcsc/*',
                    'mouser/*',
                    'search/*',
                    'settings/*'],
 'kintree.kicad': ['templates/*', 'templates_project/*']}

install_requires = \
['PySimpleGUI>=4.60.4,<5.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'digikey-api>=1.0.0,<2.0.0',
 'fuzzywuzzy>=0.18.0,<0.19.0',
 'inventree>=0.9.2,<0.10.0',
 'mouser>=0.1.2,<0.2.0',
 'multiprocess>=0.70.14,<0.71.0',
 'python-Levenshtein>=0.12.2,<0.13.0',
 'validators>=0.19.0,<0.20.0',
 'wrapt>=1.14.1,<2.0.0']

entry_points = \
{'console_scripts': ['kintree = kintree.kintree_gui:main',
                     'kintree_setup_inventree = '
                     'kintree.setup_inventree:setup_inventree']}

setup_kwargs = {
    'name': 'kintree',
    'version': '0.6.6',
    'description': 'Fast part creation in KiCad and InvenTree',
    'long_description': '# Ki-nTree\n### Fast part creation in [KiCad](https://kicad.org/) and [InvenTree](https://inventree.readthedocs.io/) \n[![License: GPL v3.0](https://img.shields.io/badge/license-GPL_v3.0-green.svg)](https://www.gnu.org/licenses/gpl-3.0)\n[![Python Versions](https://raw.githubusercontent.com/sparkmicro/Ki-nTree/main/images/python_versions.svg)](https://www.python.org/)\n[![PyPI](https://img.shields.io/pypi/v/kintree)](https://pypi.org/project/kintree/)\n[![Tests | Linting | Publishing](https://github.com/sparkmicro/Ki-nTree/actions/workflows/test_deploy.yaml/badge.svg)](https://github.com/sparkmicro/Ki-nTree/actions)\n[![Coverage Status](https://coveralls.io/repos/github/sparkmicro/Ki-nTree/badge.svg?branch=main&service=github)](https://coveralls.io/github/sparkmicro/Ki-nTree?branch=main)\n\n## Demo Videos :fast_forward: [Full Demo](https://youtu.be/haSAu926BOI) :fast_forward: [KiCad Demo](https://youtu.be/NSMfCCD0uVw)\n\n<img src="https://raw.githubusercontent.com/sparkmicro/Ki-nTree/main/images/doc/kintree_example.png"  width="auto" height="auto">\n\n## Introduction\nKi-nTree (pronounced "Key Entry" or "Key \'n\' Tree") aims to:\n* automate part creation of KiCad library parts\n* automate part creation of InvenTree parts\n* synchronize parts data between KiCad and InvenTree\n\nKi-nTree works with:\n- [Digi-Key](https://developer.digikey.com/), [Mouser](https://www.mouser.com/api-hub/), [Element14](https://partner.element14.com/docs) and [LCSC](https://lcsc.com/) **enormous** part databases and free APIs\n- the awesome open-source [Digi-Key API python library](https://github.com/peeter123/digikey-api) built and maintained by [@peeter123](https://github.com/peeter123)\n- the [Mouser Python API](https://github.com/sparkmicro/mouser-api/) built and maintained by [@eeintech](https://github.com/eeintech)\n- the awesome open-source [InvenTree Inventory Management System](https://github.com/inventree/inventree) built and maintained by [@SchrodingersGat](https://github.com/SchrodingersGat)\n- [KiCad](https://www.kicad.org/) (of course!) and their open-source [library utils](https://gitlab.com/kicad/libraries/kicad-library-utils)\n\n> **Important Note**\n>\n> Because of limited maintenance bandwidth, Ki-nTree versions `0.5.x` and newer will only support KiCad version **6**.\n> To keep using Ki-nTree with KiCad version **5**, use older `0.4.x` versions (`pip install kintree==0.4.8`).\n\nKi-nTree was developped by [@eeintech](https://github.com/eeintech) for [SPARK Microsystems](https://www.sparkmicro.com/), who generously accepted to make it open-source!\n\n## Get Started\n\n### Requirements\n\n* Ki-nTree is currently tested for Python 3.8 to 3.10 versions.\n* Ki-nTree requires a Digi-Key **production** API instance. To create one, go to https://developer.digikey.com/. Create an account, an organization and add a **production** API to your organization. Save both Client ID and Secret keys.\n> [Here is a video](https://youtu.be/OI1EGEc0Ju0) to help with the different steps\n* Ki-nTree requires a Mouser Search API key. To request one, head over to https://www.mouser.ca/api-search/ and click on "Sign Up for Search API"\n* Ki-nTree requires an Element14 Product Search API key to fetch part information for the following suppliers: Farnell (Europe), Newark (North America) and Element14 (Asia-Pacific). To request one, head over to https://partner.element14.com/ and click on "Register"\n\n### Installation (system wide)\n\n1. Install using Pip\n\n``` bash\npip install -U kintree\n```\n\n2. Run Ki-nTree\n\n``` bash\nkintree\n```\n\n### Run in virtual environment (contained)\n\n##### Linux / MacOS\n\nCreate a virtual environment and activate it with:\n\n``` bash\n$ python3 -m venv env-kintree\n$ source env-kintree/bin/activate\n```\n\nThen follow the steps from the [installation section](#installation-system-wide).\n\n##### Windows\n\nIn Git Bash, use the following commands to create and activate a virtual environment:\n``` bash\n$ python3 -m venv env-kintree\n$ source env-kintree/Scripts/activate\n```\nFor any other Windows terminal, refer to the [official documentation](https://docs.python.org/library/venv.html#creating-virtual-environments)\n\n### Packages\n#### Arch Linux\n\nKi-nTree is [available on Arch Linux\'s AUR](https://aur.archlinux.org/packages/python-kintree/) as `python-kintree`.\n\n### Usage Instructions\n\n> :warning: Warning: **KiCad library writer is still in development**: Make sure to open KiCad libraries inside the symbol editor and save them after adding symbols.\n\n#### Before Starting\n\nIf you intend to use Ki-nTree with InvenTree, this tool offers to setup the InvenTree category tree with a simple script that you can run as follow:\n\n> :warning: Warning: Before running it, make sure you have setup your category tree in your `categories.yaml` configuration file according to your own preferences, else it will use the [default setup](https://github.com/sparkmicro/Ki-nTree/blob/main/kintree/config/inventree/categories.yaml).\n\n``` bash\npython3 -m kintree.setup_inventree\n```\n\nIf the InvenTree category tree is **not setup** before starting to use Ki-nTree, you **won\'t be able to add parts** to InvenTree.\n\n#### Advanced Configuration\n\nConfiguration files are stored in the folder pointed by the `Configuration Files Folder` path in the "User Settings" window:\n\n<img src="https://raw.githubusercontent.com/sparkmicro/Ki-nTree/main/images/doc/settings_user_cache.png" width="600" height="auto">\n\n\n<details>\n<summary><b>Click here to read about configuration files</b></summary>\n<p>\n\nKi-nTree uses a number of YAML configuration files to function. New users shouldn\'t need to worry about them to try out Ki-nTree (except for `categories.yaml` as mentioned in the previous section), however they can be modified to customize behavior and workflow.\n\nBelow is a summary table of the different configuration files, their function and if they are updated by the GUI:\n| Filename | Function | GUI Update? |\n| --- | --- | --- |\n| `categories.yaml` | InvenTree categories hierarchy tree and category codes for IPN generation (see [Before Starting section](#before-starting)) | :x: |\n| `general.yaml` | General user settings | _Partial_ |\n| `internal_part_number.yaml` | Controls for IPN generation | :x: |\n| `inventree_<env>.yaml` | InvenTree login credentials, per environment (`<env>=[\'dev\', \'prod\']`) | :heavy_check_mark: |\n| `kicad.yaml` | KiCad symbol, footprint and library paths | :heavy_check_mark: |\n| `kicad_map.yaml` | Mapping between InvenTree parent categories and KiCad symbol/footprint libraries and templates | :x: |\n| `parameters.yaml` | List of InvenTree parameters templates (see [InvenTree Part Parameters documentation](https://docs.inventree.org/en/latest/part/parameter/)) | :x: |\n| `parameters_filters.yaml` | Mapping between InvenTree parent categories and InvenTree parameters templates | :x: |\n| `search_api.yaml` | Generic controls for Supplier search APIs like cache validity and category matching ratio | :x: |\n| `supplier_parameters.yaml` | Mapping between InvenTree parameters templates and suppliers parameters/attributes, sorted by InvenTree parent categories (see [Part Parameters section](#part-parameters)) | :x: |\n| `<supplier>_config.yaml` | Mapping for supplier name and search results fields, to overwrite defaults (`<supplier>=[\'digikey\', \'element14\', \'lcsc\', \'mouser\']`) | :x: |\n| `<supplier>_api.yaml` | Required supplier API fields, custom to each supplier (`<supplier>=[\'digikey\', \'element14\', \'lcsc\', \'mouser\']`) | :heavy_check_mark: |\n| `digikey_categories.yaml` | Mapping between InvenTree categories and Digi-Key categories | :heavy_check_mark: |\n| `digikey_parameters.yaml` | Mapping between InvenTree parameters and Digi-Key parameters/attributes | :x: |\n\n> In versions `0.6.x` and older, Ki-nTree only supports matching between InvenTree and Digi-Key categories and parameters/attibutes (help wanted!)\n\n</p>\n</details>\n\n#### InvenTree Permissions\n\nEach InvenTree user has a set of permissions associated to them.\nPlease refer to the [InvenTree documentation](https://inventree.readthedocs.io/en/latest/settings/permissions/) to understand how to setup user permissions.\n\nThe minimum set of user/group permission required to add parts to InvenTree is:\n- "Part - Add" to add InvenTree parts\n- "Purchase Orders - Add" to add manufacturers, suppliers and their associated parts\n\nIf you wish to automatically add subcategories while creating InvenTree parts, you will need to enable the "Part Categories - Add" permission.\n\nNote that each time you enable the "Add" permission to an object, InvenTree automatically enables the "Change" permission to that object too.\n\n#### Settings\n1. With Ki-nTree GUI open, click on "Settings > Supplier > Digi-Key" and fill in both Digi-Key API Client ID and Secret keys (optional: click on "Test" to [get an API token](#get-digi-key-api-token))\n2. Click on "Settings > Supplier > Mouser" and fill in the Mouser part search API key\n3. Click on "Settings > Supplier > Element14" and fill in the Element14 product search API key (key is shared with Farnell and Newark)\n4. Click on "Settings > KiCad", browse to the location where KiCad symbol, template and footprint libraries are stored on your computer then click "Save"\n5. If you intend to use InvenTree with this tool, click on "Settings > InvenTree" and fill in your InvenTree server address and credentials then click "Save" (optional: click on "Test" to get an API token)\n\n#### Get Digi-Key API token\n<details>\n<summary>Show steps (click to expand)</summary>\n<p>\n\nEnter your Digi-Key developper account credentials then login. The following page will appear (`user@email.com` will show your email address):\n\n<img src="https://raw.githubusercontent.com/sparkmicro/Ki-nTree/main/images/doc/digikey_api_approval_request.png" width="600" height="auto">\n\nClick on "Allow", another page will open.  \nClick on the "Advanced" button, then click on "Proceed to localhost (unsafe)" at the bottom of the page:\n\n<img src="https://raw.githubusercontent.com/sparkmicro/Ki-nTree/main/images/doc/digikey_api_approval_request2.png"  width="600" height="auto">\n\n> On Chrome, if the "Proceed to localhost (unsafe)" link does not appear, enable the following flag: [chrome://flags/#allow-insecure-localhost](chrome://flags/#allow-insecure-localhost)\n\nLastly, a new page will open with a "You may now close this window." message, proceed to get the token.\n\n</p>\n</details>\n\n#### Part Parameters\n\nKi-nTree uses **supplier** parameters to populate **InvenTree** parameters. In order to match between supplier and InvenTree, users need to setup the configuration file `supplier_parameters.yaml` with the following mapping for each category:\n``` yaml\nCATEGORY_NAME:\n  INVENTREE_PARAMETER_NAME:\n    - SUPPLIER_1_PARAMETER_NAME_1\n    - SUPPLIER_1_PARAMETER_NAME_2\n    - SUPPLIER_2_PARAMETER_NAME_1\n```\n\nRefer to [this file](https://github.com/sparkmicro/Ki-nTree/blob/main/kintree/config/search/supplier_parameters.yaml) as a starting point / example.\n\n#### Part Number Search\n\nKi-nTree currently supports APIs for the following electronics suppliers: Digi-Key, Mouser, Element14 and LCSC.\n\n1. In the main window, enter the part number and select the supplier in drop-down list, then click "CREATE". It will start by fetching part data using the supplier\'s API\n2. In the case Digi-Key has been selected and the API token is not found or expired, a browser window will pop-up. To get a new token: [follow those steps](#get-digi-key-api-token)\n3. Once the part data has been successfully fetched from the supplier\'s API, you will be prompted to add/confirm/edit the part information, followed by the `Category` and `Subcategory` to use for this part (Ki-nTree tries to match them automatically)\n4. Then, you will be prompted with selecting the KiCad symbol library, the template and the footprint library to use for this part\n5. It will take some time to complete the part creation in InvenTree and/or KiCad, once it finishes you\'ll be notified of the result  \n6. Finally, if the part was created or found in InvenTree, your browser will automatically open a new tab with the part information\n\nEnjoy!\n\n*For any problem/bug you find, please [report an issue](https://github.com/sparkmicro/Ki-nTree/issues).*\n\n## Development\n\n### Requirements\n\nYou need `python>=3.8` and `poetry`.\n\nYou can install poetry by following the instructions [on its official website](https://python-poetry.org/docs/master/#installation), by using `pip install poetry` or by installing a package on your Linux distro.\n\n### Setup and run\n1. Clone this repository\n``` bash\ngit clone https://github.com/sparkmicro/Ki-nTree\n```\n\n2. Install the requirements into a `poetry`-managed virtual environment\n``` bash\npoetry install\nInstalling dependencies from lock file\n...\nInstalling the current project: kintree (0.4.99)\n```\n> Note: the version is not accurate (placeholder only)\n\n3. Run Ki-nTree in the virtual environment\n```bash\npoetry run python -m kintree_gui\n```\nor\n\n```bash\n$ poetry shell\n$ python -m kintree_gui\n```\n\n#### Build\n1. Make sure you followed the previous installation steps, then run:\n``` bash\n$ poetry build\nBuilding kintree (0.6.99)\n  - Building sdist\n  - Built kintree-0.6.99.tar.gz\n  - Building wheel\n  - Built kintree-0.6.99-py3-none-any.whl\n```\n2. Exit the virtual environment (`Ctrl + D` on Linux; you can also close the\n   terminal and reopen it in the same folder).\n\n   Run `pip install dist/<wheel_file>.whl` with the file name from the previous\n   step. For example:\n\n```bash\npip install dist/kintree-0.6.99-py3-none-any.whl\n```\n\n3. You can now start Ki-nTree by typing `kintree` in the terminal, provided\n   that your python dist path is a part of your `$PATH`.\n\n## Roadmap\n\n> Priority goes to implementing new features over GUI improvements  \n> Open to new ideas and pull requests :smiley:\n\n#### Versions 1.0.x or later\n##### New Features\n\n- Revamp category selection based on hierarchical structure from InvenTree ([reference](https://github.com/sparkmicro/Ki-nTree/issues/87)) \n- Allow user to decide the category code to use for IPN\n- Enable option to download and save PDF files locally/to internal server storage\n\n##### Improvements\n\n- Optimize the category search ([reference](https://github.com/sparkmicro/Ki-nTree/issues/104))\n- Migrate KiCad library management to [KiUtils](https://github.com/mvnmgrx/kiutils) ([reference](https://github.com/sparkmicro/Ki-nTree/issues/119))\n\n##### GUI\n\n- Migrate to [Flet](https://github.com/flet-dev/flet) ([reference](https://github.com/sparkmicro/Ki-nTree/issues/37))\n\n## License\n\nThe Ki-nTree source code is licensed under the [GPL3.0 license](https://github.com/sparkmicro/Ki-nTree/blob/main/LICENSE) as it uses source code under that license:\n* https://gitlab.com/kicad/libraries/kicad-library-utils\n* https://github.com/peeter123/digikey-api\n\nThe [KiCad templates](https://github.com/sparkmicro/Ki-nTree/tree/main/kintree/kicad/templates) are licensed under the [Creative Commons CC0 1.0 license](https://github.com/sparkmicro/Ki-nTree/blob/main/kintree/kicad/templates/LICENSE) which means that "you can copy, modify, distribute and perform the work, even for commercial purposes, all without asking permission" ([reference](https://creativecommons.org/publicdomain/zero/1.0/)).',
    'author': 'eeintech',
    'author_email': 'eeintech@eeinte.ch',
    'maintainer': 'eeintech',
    'maintainer_email': 'eeintech@eeinte.ch',
    'url': 'https://github.com/sparkmicro/Ki-nTree',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



""" DISCLAIMER
kicad_sym classes and methods were directly imported from: https://gitlab.com/kicad/libraries/kicad-library-utils
Last import date: Jan 3rd, 2023
"""

"""
Library for processing KiCad's symbol files.
"""

import json
import math
import re
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from . import sexpr


class KicadFileFormatError(ValueError):
    """any kind of problem discovered while parsing a KiCad file"""


# KiCad can only handle multiples of 90 degrees
VALID_ROTATIONS = frozenset({0, 90, 180, 270})


def mil_to_mm(mil: float) -> float:
    return round(mil * 0.0254, 6)


def mm_to_mil(mm: float) -> float:
    return round(mm / 0.0254)


def _parse_at(i):
    sexpr_at = _get_array(i, "at")[0]
    posx = sexpr_at[1]
    posy = sexpr_at[2]
    if len(sexpr_at) == 4:
        rot = sexpr_at[3]
    else:
        rot = None
    return (posx, posy, rot)


def _get_array(
    data,
    value,
    result: Optional[List[Any]] = None,
    level: int = 0,
    max_level: Optional[int] = None,
):
    """return the array which has value as first element"""
    if result is None:
        result = []

    if max_level is not None and max_level <= level:
        return result

    level += 1

    for i in data:
        if isinstance(i, list):
            _get_array(i, value, result, level=level, max_level=max_level)
        else:
            if i == value:
                result.append(data)
    return result


def _get_array2(data, value):
    ret = []
    for i in data:
        if isinstance(i, list) and i[0] == value:
            ret.append(i)
    return ret


def _get_color(sexpr) -> Optional["Color"]:
    col = None
    for i in sexpr:
        if isinstance(i, list) and i[0] == "color":
            col = Color(i[1], i[2], i[3], i[4])
    return col


def _get_stroke(sexpr) -> Tuple[Optional[int], Optional["Color"]]:
    width = None
    col = None
    for i in sexpr:
        if isinstance(i, list) and i[0] == "stroke":
            width = _get_value_of(i, "width")
            col = _get_color(i)
            break
    return (width, col)


def _get_fill(sexpr) -> Tuple[Optional[Any], Optional["Color"]]:
    fill = None
    col = None
    for i in sexpr:
        if isinstance(i, list) and i[0] == "fill":
            fill = _get_value_of(i, "type")
            col = _get_color(i)
            break
    return (fill, col)


def _get_xy(sexpr, lookup) -> Tuple[float, float]:
    for i in sexpr:
        if isinstance(i, list) and i[0] == lookup:
            return (i[1], i[2])
    return (0.0, 0.0)


def _get_value_ofRecursively(data, path, item_to_get=False):
    """return the array which has value as first element, but recursively

    if item_to_get is != 0, return the array element with that index
    """
    # if we walked the whole path we are done. return the data
    if not path:
        # in some cases it is useful to only get the 2nd item, for
        # example ['lenght', 42] should just return 42
        if item_to_get != 0:
            return data[item_to_get]
        return data

    for i in data:
        # look at sub-arrays, if their first element matches the path-spec,
        # strip the front item from the path list and do this recursively
        if isinstance(i, list) and i[0] == path[0]:
            return _get_value_ofRecursively(i, path[1:], item_to_get)


def _get_value_of(data, lookup, default=None):
    """find the array which has lookup as first element, return its 2nd element"""
    for i in data:
        if isinstance(i, list) and i[0] == lookup:
            return i[1]
    return default


def _has_value(data, lookup) -> bool:
    """return true if the lookup item exists"""
    for i in data:
        if isinstance(i, list) and i[0] == lookup:
            return True
    return False


class KicadSymbolBase:
    def as_json(self):
        return json.dumps(self, default=lambda x: x.__dict__, indent=2)

    def compare_pos(self, x, y):
        if hasattr(self, "posx") and hasattr(self, "posy"):
            return round(self.posx, 6) == round(x, 6) and round(self.posy, 6) == round(
                y, 6
            )
        return False

    def is_unit(self, unit, demorgan):
        if hasattr(self, "unit") and hasattr(self, "demorgan"):
            return self.unit == unit and self.demorgan == demorgan
        return False

    @classmethod
    def quoted_string(cls, s: str) -> str:
        s = re.sub(r"\n", r"\\n", s)
        return '"' + s + '"'

    @classmethod
    def dir_to_rotation(cls, d: str) -> int:
        if d == "R":
            return 0
        elif d == "U":
            return 90
        elif d == "L":
            return 180
        elif d == "D":
            return 270
        else:
            raise ValueError(
                f"Invalid direction requested: {d} (should be one of: R / U / L / D"
            )


@dataclass
class Color(KicadSymbolBase):
    """Encode the color of an entiry. Currently not used in the kicad_sym format"""

    r: int
    g: int
    b: int
    a: int


@dataclass
class TextEffect(KicadSymbolBase):
    """Encode the text effect of an entiry"""

    sizex: float
    sizey: float
    is_italic: bool = False
    is_bold: bool = False
    is_hidden: bool = False
    is_mirrored: bool = False
    h_justify: str = "center"
    v_justify: str = "center"
    color: Optional[Color] = None
    font: Optional[str] = None

    @classmethod
    def new_mil(cls, size: float) -> "TextEffect":
        return cls(mil_to_mm(size), mil_to_mm(size))

    def get_sexpr(self):
        fnt = ["font", ["size", self.sizex, self.sizey]]
        if self.is_italic:
            fnt.append("italic")
        if self.is_bold:
            fnt.append("bold")
        sx = ["effects", fnt]
        if self.is_mirrored:
            sx.append("mirror")
        if self.color:
            sx.append(self.color.get_sexpr())
        if self.is_hidden:
            sx.append("hide")

        justify = ["justify"]
        if self.h_justify and self.h_justify != "center":
            justify.append(self.h_justify)
        if self.v_justify and self.v_justify != "center":
            justify.append(self.v_justify)

        if len(justify) > 1:
            sx.append(justify)
        return sx

    @classmethod
    def from_sexpr(cls, sexpr):
        if sexpr.pop(0) != "effects":
            return None
        font = _get_array(sexpr, "font")[0]
        (sizex, sizey) = _get_xy(font, "size")
        is_italic = "italic" in font
        is_bold = "bold" in font
        is_hidden = "hide" in sexpr
        is_mirrored = "mirror" in sexpr
        justify = _get_array2(sexpr, "justify")
        h_justify = "center"
        v_justify = "center"
        if justify:
            if "top" in justify[0]:
                v_justify = "top"
            if "bottom" in justify[0]:
                v_justify = "bottom"
            if "left" in justify[0]:
                h_justify = "left"
            if "right" in justify[0]:
                h_justify = "right"
        return cls(
            sizex,
            sizey,
            is_italic,
            is_bold,
            is_hidden,
            is_mirrored,
            h_justify,
            v_justify,
        )


@dataclass
class AltFunction(KicadSymbolBase):
    name: str
    etype: str
    shape: str = "line"

    def get_sexpr(self) -> List[str]:
        return ["alternate", self.quoted_string(self.name), self.etype, self.shape]

    @classmethod
    def from_sexpr(cls, sexpr) -> "AltFunction":
        (identifier, name, etype, shape) = sexpr
        return AltFunction(name, etype, shape)


@dataclass
class Pin(KicadSymbolBase):
    name: str
    number: str
    etype: str
    posx: float = 0.0
    posy: float = 0.0
    rotation: int = 0
    shape: str = "line"
    length: float = 2.54
    is_global: bool = False
    is_hidden: bool = False
    number_int: Optional[int] = None
    name_effect: Optional[TextEffect] = None
    number_effect: Optional[TextEffect] = None
    altfuncs: List[AltFunction] = field(default_factory=list)
    unit: int = 0
    demorgan: int = 0

    def __post_init__(self):
        # try to parse the pin_number into an integer if possible
        if self.number_int is None and re.match(r"^\d+$", self.number):
            self.number_int = int(self.number)
        # There is some weird thing going on with the instance creation of name_effect and
        # number_effect.
        # when creating lots of pins from scratch, the id() of their name_effect member is the same
        # that is most likely the result of some optimization
        # to circumvent that, we create instances explicitly
        if self.name_effect is None:
            self.name_effect = TextEffect(1.27, 1.27)
        if self.number_effect is None:
            self.number_effect = TextEffect(1.27, 1.27)

    @classmethod
    def _parse_name_or_number(cls, i, typ="name"):
        """Convert a sexpr pin-name or pin-number into a python dict"""
        sexpr_n = _get_array(i, typ)[0]
        name = sexpr_n[1]
        effects = TextEffect.from_sexpr(_get_array(sexpr_n, "effects")[0])
        return (name, effects)

    def get_sexpr(self):
        sx = [
            "pin",
            self.etype,
            self.shape,
            ["at", self.posx, self.posy, self.rotation],
        ]
        if self.is_global:
            sx.append("global")
        sx.append(["length", self.length])
        if self.is_hidden:
            sx.append("hide")
        sx.append(["name", self.quoted_string(self.name), self.name_effect.get_sexpr()])
        sx.append(
            ["number", self.quoted_string(self.number), self.number_effect.get_sexpr()]
        )
        for altfn in self.altfuncs:
            sx.append(altfn.get_sexpr())

        return sx

    def get_direction(self) -> str:
        if self.rotation == 0:
            return "R"
        elif self.rotation == 90:
            return "U"
        elif self.rotation == 180:
            return "L"
        elif self.rotation == 270:
            return "D"
        else:
            raise NotImplementedError(f"Invalid 'rotation' of Pin: {self.rotation}")

    def is_duplicate(self, p: "Pin") -> bool:
        if (
            p.number == self.number
            and p.unit == self.unit
            and p.demorgan == self.demorgan
        ):
            return True
        return False

    @classmethod
    def from_sexpr(cls, sexpr, unit: int, demorgan: int) -> "Pin":
        is_global = False
        # The first 3 items are pin, type and shape
        if sexpr.pop(0) != "pin":
            return None
        etype = sexpr.pop(0)
        shape = sexpr.pop(0)
        # the 4th item (global) is optional
        if sexpr[0] == "global":
            sexpr.pop(0)
            is_global = True
        # fetch more properties
        is_hidden = "hide" in sexpr
        length = _get_value_of(sexpr, "length")
        (posx, posy, rotation) = _parse_at(sexpr)
        (name, name_effect) = cls._parse_name_or_number(sexpr)
        (number, number_effect) = cls._parse_name_or_number(sexpr, typ="number")

        if rotation not in VALID_ROTATIONS:
            raise ValueError(
                f"Invalid 'rotation' attribute value for pin: {rotation}"
                f" (must be one of {set(VALID_ROTATIONS)})"
            )
        altfuncs = []
        alt_n = _get_array(sexpr, "alternate")
        for alt_sexpr in alt_n:
            altfuncs.append(AltFunction.from_sexpr(alt_sexpr))
        # we also need the pin-number as integer, try to convert it.
        # Some pins won't work since they are called 'MP' or similar
        number_int = None
        try:
            number_int = int(number)
        except ValueError:
            pass
        # create and return a pin with the extracted values
        return Pin(
            name,
            number,
            etype,
            posx,
            posy,
            rotation,
            shape,
            length,
            is_global,
            is_hidden,
            number_int,
            name_effect,
            number_effect,
            altfuncs=altfuncs,
            unit=unit,
            demorgan=demorgan,
        )


@dataclass
class Circle(KicadSymbolBase):
    centerx: float
    centery: float
    radius: float
    stroke_width: float = 0.254
    stroke_color: Optional[Color] = None
    fill_type: str = "none"
    fill_color: Optional[Color] = None
    unit: int = 0
    demorgan: int = 0

    def get_sexpr(self) -> List[Any]:
        sx = [
            "circle",
            ["center", self.centerx, self.centery],
            ["radius", self.radius],
            ["stroke", ["width", self.stroke_width]],
            ["fill", ["type", self.fill_type]],
        ]
        return sx

    @classmethod
    def from_sexpr(cls, sexpr, unit: int, demorgan: int):
        # The first 3 items are pin, type and shape
        if sexpr.pop(0) != "circle":
            return None
        # the 1st element
        (centerx, centery) = _get_xy(sexpr, "center")
        radius = _get_value_of(sexpr, "radius")
        (stroke, scolor) = _get_stroke(sexpr)
        (fill, fcolor) = _get_fill(sexpr)
        return Circle(
            centerx,
            centery,
            radius,
            stroke,
            scolor,
            fill,
            fcolor,
            unit=unit,
            demorgan=demorgan,
        )


@dataclass
class Arc(KicadSymbolBase):
    # (arc (start -5.08 -1.27) (mid 0 -6.35) (end 5.08 -1.27)
    startx: float
    starty: float
    endx: float
    endy: float
    midx: float
    midy: float
    stroke_width: float = 0.254
    stroke_color: Optional[Color] = None
    fill_type: str = "none"
    fill_color: Optional[Color] = None
    unit: int = 0
    demorgan: int = 0

    def get_sexpr(self) -> List[Any]:
        sx = [
            "arc",
            ["start", self.startx, self.starty],
            ["mid", self.midx, self.midy],
            ["end", self.endx, self.endy],
        ]
        sx.append(["stroke", ["width", self.stroke_width]])
        sx.append(["fill", ["type", self.fill_type]])
        return sx

    @classmethod
    def from_sexpr(cls, sexpr, unit: int, demorgan: int) -> Optional["Arc"]:
        if sexpr.pop(0) != "arc":
            return None
        (startx, starty) = _get_xy(sexpr, "start")
        (endx, endy) = _get_xy(sexpr, "end")
        (midx, midy) = _get_xy(sexpr, "mid")
        (stroke, scolor) = _get_stroke(sexpr)
        (fill, fcolor) = _get_fill(sexpr)
        return Arc(
            startx,
            starty,
            endx,
            endy,
            midx,
            midy,
            stroke,
            scolor,
            fill,
            fcolor,
            unit=unit,
            demorgan=demorgan,
        )


@dataclass
class Point(KicadSymbolBase):
    x: float
    y: float

    @classmethod
    def new_mil(cls, x: float, y: float) -> "Point":
        return cls(mil_to_mm(x), mil_to_mm(y))

    def get_sexpr(self):
        return ["xy", self.x, self.y]


@dataclass
class Polyline(KicadSymbolBase):
    points: List[Point]
    stroke_width: float = 0.254
    stroke_color: Optional[Color] = None
    fill_type: str = "none"
    fill_color: Optional[Color] = None
    unit: int = 0
    demorgan: int = 0

    def get_sexpr(self):
        pts_list = [x.get_sexpr() for x in self.points]
        pts_list.insert(0, "pts")
        sx = [
            "polyline",
            pts_list,
            ["stroke", ["width", self.stroke_width]],
            ["fill", ["type", self.fill_type]],
        ]
        return sx

    def is_closed(self) -> bool:
        # if the last and first point are the same, we consider the polyline closed
        # a closed triangle will have 4 points (A-B-C-A) stored in the list of points
        return (len(self.points) > 3) and (self.points[0] == self.points[-1])

    def get_boundingbox(self) -> Tuple[float, float, float, float]:
        if self.points:
            minx = min(p.x for p in self.points)
            maxx = max(p.x for p in self.points)
            miny = min(p.y for p in self.points)
            maxy = max(p.y for p in self.points)
            return (maxx, maxy, minx, miny)
        else:
            return (0, 0, 0, 0)

    def as_rectangle(self) -> "Rectangle":
        (maxx, maxy, minx, miny) = self.get_boundingbox()
        return Rectangle(
            minx,
            maxy,
            maxx,
            miny,
            self.stroke_width,
            self.stroke_color,
            self.fill_type,
            self.fill_color,
            unit=self.unit,
            demorgan=self.demorgan,
        )

    def get_center_of_boundingbox(self) -> Tuple[float, float]:
        (maxx, maxy, minx, miny) = self.get_boundingbox()
        return ((minx + maxx) / 2, ((miny + maxy) / 2))

    def is_rectangle(self) -> bool:
        # a rectangle has 5 points and is closed
        if len(self.points) != 5 or not self.is_closed():
            return False

        # construct lines between the points
        p0 = self.points[0]
        for p1_idx in range(1, len(self.points)):
            p1 = self.points[p1_idx]
            dx = p1.x - p0.x
            dy = p1.y - p0.y
            if dx != 0 and dy != 0:
                # if a line is neither horizontal or vertical its not
                # part of a rectangle
                return False
            # select next point
            p0 = p1

        return True

    @classmethod
    def from_sexpr(cls, sexpr, unit: int, demorgan: int) -> Optional["Polyline"]:
        pts = []
        if sexpr.pop(0) != "polyline":
            return None
        for p in _get_array(sexpr, "pts")[0]:
            if "xy" in p:
                pts.append(Point(p[1], p[2]))

        (stroke, scolor) = _get_stroke(sexpr)
        (fill, fcolor) = _get_fill(sexpr)
        return Polyline(pts, stroke, scolor, fill, fcolor, unit=unit, demorgan=demorgan)


@dataclass
class Text(KicadSymbolBase):
    text: str
    posx: float
    posy: float
    rotation: float
    effects: TextEffect
    unit: int = 0
    demorgan: int = 0

    def get_sexpr(self) -> List[Any]:
        sx = [
            "text",
            self.quoted_string(self.text),
            ["at", self.posx, self.posy, self.rotation],
            self.effects.get_sexpr(),
        ]
        return sx

    @classmethod
    def from_sexpr(cls, sexpr, unit: int, demorgan: int) -> Optional["Text"]:
        if sexpr.pop(0) != "text":
            return None
        text = sexpr.pop(0)
        (posx, posy, rotation) = _parse_at(sexpr)
        effects = TextEffect.from_sexpr(_get_array(sexpr, "effects")[0])
        return Text(text, posx, posy, rotation, effects, unit=unit, demorgan=demorgan)


@dataclass
class Rectangle(KicadSymbolBase):
    """
    Some v6 symbols use rectangles, newer ones encode them as polylines.
    At some point in time we can most likely remove this class since its not used anymore
    """

    startx: float
    starty: float
    endx: float
    endy: float
    stroke_width: float = 0.254
    stroke_color: Optional[Color] = None
    fill_type: str = "background"
    fill_color: Optional[Color] = None
    unit: int = 0
    demorgan: int = 0

    @classmethod
    def new_mil(
        cls, sx: float, sy: float, ex: float, ey: float, fill: str = "background"
    ) -> "Rectangle":
        r = cls(mil_to_mm(sx), mil_to_mm(sy), mil_to_mm(ex), mil_to_mm(ey))
        if fill in ["none", "outline", "background"]:
            r.fill_type = fill
        return r

    def get_sexpr(self) -> List[Any]:
        sx = [
            "rectangle",
            ["start", self.startx, self.starty],
            ["end", self.endx, self.endy],
            ["stroke", ["width", self.stroke_width]],
            ["fill", ["type", self.fill_type]],
        ]
        return sx

    def as_polyline(self) -> Polyline:
        pts = [
            Point(self.startx, self.starty),
            Point(self.endx, self.starty),
            Point(self.endx, self.endy),
            Point(self.startx, self.endy),
            Point(self.startx, self.starty),
        ]
        return Polyline(
            pts,
            self.stroke_width,
            self.stroke_color,
            self.fill_type,
            self.fill_color,
            unit=self.unit,
            demorgan=self.demorgan,
        )

    def get_center(self) -> Tuple[float, float]:
        x = (self.endx + self.startx) / 2.0
        y = (self.endy + self.starty) / 2.0
        return (x, y)

    @classmethod
    def from_sexpr(cls, sexpr, unit: int, demorgan: int) -> Optional["Rectangle"]:
        if sexpr.pop(0) != "rectangle":
            return None
        # the 1st element
        (startx, starty) = _get_xy(sexpr, "start")
        (endx, endy) = _get_xy(sexpr, "end")
        (stroke, scolor) = _get_stroke(sexpr)
        (fill, fcolor) = _get_fill(sexpr)
        return Rectangle(
            startx,
            starty,
            endx,
            endy,
            stroke,
            scolor,
            fill,
            fcolor,
            unit=unit,
            demorgan=demorgan,
        )


@dataclass
class Property(KicadSymbolBase):
    name: str
    value: str
    idd: int
    posx: float = 0.0
    posy: float = 0.0
    rotation: float = 0.0
    effects: Optional[TextEffect] = None

    def __post_init__(self):
        # There is some weird thing going on with the instance creation of effect.
        # Do the same trick as we do with Pin().
        if self.effects is None:
            self.effects = TextEffect(1.27, 1.27)

    def get_sexpr(self) -> List[Any]:
        sx = [
            "property",
            self.quoted_string(self.name),
            self.quoted_string(self.value),
            ["id", self.idd],
            ["at", self.posx, self.posy, self.rotation],
            self.effects.get_sexpr(),
        ]
        return sx

    def set_pos_mil(self, x: float, y: float, rot: int = 0) -> None:
        self.posx = mil_to_mm(x)
        self.posy = mil_to_mm(y)
        # TODO: verify whether we should really (silently) restrict the rotation of all properties
        if rot in VALID_ROTATIONS:
            self.rotation = rot

    @classmethod
    def from_sexpr(cls, sexpr, unit: int = 0) -> Optional["Property"]:
        if sexpr.pop(0) != "property":
            return None
        name = sexpr.pop(0)
        value = sexpr.pop(0)
        idd = _get_value_of(sexpr, "id")
        (posx, posy, rotation) = _parse_at(sexpr)
        effects = TextEffect.from_sexpr(_get_array(sexpr, "effects")[0])
        return Property(name, value, idd, posx, posy, rotation, effects)


@dataclass
class KicadSymbol(KicadSymbolBase):
    name: str
    libname: str
    filename: str = field(compare=False)
    properties: List[Property] = field(default_factory=list)
    pins: List[Pin] = field(default_factory=list)
    rectangles: List[Rectangle] = field(default_factory=list)
    circles: List[Circle] = field(default_factory=list)
    arcs: List[Arc] = field(default_factory=list)
    polylines: List[Polyline] = field(default_factory=list)
    texts: List[Text] = field(default_factory=list)
    pin_names_offset: float = 0.508
    hide_pin_names: bool = False
    hide_pin_numbers: bool = False
    is_power: bool = False
    in_bom: bool = True
    on_board: bool = True
    extends: Optional[str] = None
    unit_count: int = 0
    demorgan_count: int = 0

    def __post_init__(self):
        if self.filename == "":
            raise ValueError("Filename can not be empty")
        self.libname = Path(self.filename).stem

    def get_sexpr(self) -> List[str]:
        # add header
        full_name = self.quoted_string("{}".format(self.name))
        sx = ["symbol", full_name]
        if self.extends:
            sx.append(["extends", self.quoted_string(self.extends)])

        pn = ["pin_names"]
        if self.pin_names_offset != 0.508:
            pn.append(["offset", self.pin_names_offset])
        if self.hide_pin_names:
            pn.append("hide")
        if len(pn) > 1:
            sx.append(pn)

        sx.append(["in_bom", "yes" if self.in_bom else "no"])
        sx.append(["on_board", "yes" if self.on_board else "no"])
        if self.is_power:
            sx.append(["power"])
        if self.hide_pin_numbers:
            sx.append(["pin_numbers", "hide"])

        # add properties
        for prop in self.properties:
            sx.append(prop.get_sexpr())

        # add units
        for d in range(0, self.demorgan_count + 1):
            for u in range(0, self.unit_count + 1):
                hdr = self.quoted_string("{}_{}_{}".format(self.name, u, d))
                sx_i = ["symbol", hdr]
                for pin in (
                    self.arcs
                    + self.circles
                    + self.texts
                    + self.rectangles
                    + self.polylines
                    + self.pins
                ):
                    if pin.is_unit(u, d):
                        sx_i.append(pin.get_sexpr())

                if len(sx_i) > 2:
                    sx.append(sx_i)

        return sx

    def get_center_rectangle(self, units: List[int]) -> Optional[Polyline]:
        # return a polyline for the requested unit that is a rectangle
        # and is closest to the center
        candidates = {}
        # building a dict with floats as keys.. there needs to be a rule against that^^
        pl_rects = [i.as_polyline() for i in self.rectangles]
        pl_rects.extend(pl for pl in self.polylines if pl.is_rectangle())
        for pl in pl_rects:
            if pl.unit in units:
                # extract the center, calculate the distance to origin
                (x, y) = pl.get_center_of_boundingbox()
                dist = math.sqrt(x * x + y * y)
                candidates[dist] = pl

        if candidates:
            # sort the list return the first (smallest) item
            return candidates[sorted(candidates.keys())[0]]
        return None

    def get_pinstacks(self) -> Dict[str, List[Pin]]:
        stacks = {}
        for pin in self.pins:
            # if the unit is 0 that means this pin is common to all units
            unit_list = [pin.unit]
            if pin.unit == 0:
                unit_list = list(range(1, self.unit_count + 1))

            # if the unit is 0 that means this pin is common to all units
            demorgan_list = [pin.demorgan]
            if pin.demorgan == 0:
                demorgan_list = list(range(1, self.demorgan_count + 1))

            # add the pin to the correct stack
            for demorgan in demorgan_list:
                for unit in unit_list:
                    loc = "x{0}_y{1}_u{2}_d{3}".format(
                        pin.posx, pin.posy, unit, demorgan
                    )
                    if loc in stacks:
                        stacks[loc].append(pin)
                    else:
                        stacks[loc] = [pin]
        return stacks

    def get_property(self, pname: str) -> Optional[Property]:
        for p in self.properties:
            if p.name == pname:
                return p

        return None

    def add_default_properties(self) -> None:
        defaults = [
            {"i": 0, "n": "Reference", "v": "U", "h": False},
            {"i": 1, "n": "Value", "v": self.name, "h": False},
            {"i": 2, "n": "Footprint", "v": "", "h": True},
            {"i": 3, "n": "Datasheet", "v": "", "h": True},
            {"i": 4, "n": "ki_locked", "v": "", "h": True},
            {"i": 5, "n": "ki_keywords", "v": "", "h": True},
            {"i": 6, "n": "ki_description", "v": "", "h": True},
            {"i": 7, "n": "ki_fp_filters", "v": "", "h": False},
        ]

        for prop in defaults:
            if self.get_property(prop["n"]) is None:
                p = Property(prop["n"], prop["v"], prop["i"])
                p.effects.is_hidden = prop["h"]
                self.properties.append(p)

    @classmethod
    def new(
        cls,
        name: str,
        libname: str,
        reference: str = "U",
        footprint: str = "",
        datasheet: str = "",
        keywords: str = "",
        description: str = "",
        fp_filters: str = "",
    ):
        sym = cls(name, libname, libname + ".kicad_sym")
        sym.add_default_properties()
        sym.get_property("Reference").value = reference
        sym.get_property("Footprint").value = footprint
        sym.get_property("Datasheet").value = datasheet
        sym.get_property("ki_keywords").value = keywords
        sym.get_property("ki_description").value = description
        if isinstance(fp_filters, list):
            fp_filters = " ".join(fp_filters)
        sym.get_property("ki_fp_filters").value = fp_filters
        return sym

    def get_fp_filters(self) -> List[str]:
        filters = self.get_property("ki_fp_filters")
        if filters:
            return filters.value.split(" ")
        else:
            return []

    def is_graphic_symbol(self) -> bool:
        return self.extends is None and (
            not self.pins or self.get_property("Reference").value == "#SYM"
        )

    def is_power_symbol(self) -> bool:
        return self.is_power

    def is_locked(self) -> bool:
        return self.get_property("ki_locked") is not None

    def get_pins_by_name(self, name: str) -> List[Pin]:
        pins = []
        for pin in self.pins:
            if pin.name == name:
                pins += [pin]
        return pins

    def get_pins_by_number(self, num) -> Optional[Pin]:
        for pin in self.pins:
            # @todo num does not exist in pin...
            if pin.num == str(num):
                return pin
        return None

    def filter_pins(
        self,
        name: Optional[str] = None,
        direction: Optional[str] = None,
        electrical_type: Optional[str] = None,
    ) -> List[Pin]:
        pins = []
        for pin in self.pins:
            if (
                (name and pin.name == name)
                or (direction and pin.rotation == self.dir_to_rotation(direction))
                or (electrical_type and pin.etype == electrical_type)
            ):
                pins.append(pin)
        return pins

    # Heuristics, which tries to determine whether this is a "small" component (resistor,
    # capacitor, LED, diode, transistor, ...).
    def is_small_component_heuristics(self) -> bool:
        if len(self.pins) <= 2:
            return True

        filled_rect = self.get_center_rectangle(range(self.unit_count))

        # if there is no filled rectangle as symbol outline and we have 3 or 4 pins, we assume this
        # is a small symbol
        if (3 <= len(self.pins) <= 4) and (filled_rect is None):
            return True

        return False


@dataclass
class KicadLibrary(KicadSymbolBase):
    """
    A class to parse kicad_sym files format of the KiCad
    """

    filename: str
    symbols: List[KicadSymbol] = field(default_factory=list)
    generator: str = "kicad-library-utils"
    version: str = "20211014"

    def write(self) -> None:
        lib_file = open(self.filename, "w")
        lib_file.write(self.get_sexpr())
        lib_file.close()

    def get_sexpr(self) -> str:
        sx = [
            "kicad_symbol_lib",
            ["version", self.version],
            ["generator", self.generator],
        ]
        for sym in self.symbols:
            sx.append(sym.get_sexpr())
        return sexpr.build_sexp(sx)

    @classmethod
    def from_file(cls, filename: str) -> "KicadLibrary":
        """
        Parse a symbol library from a file.

        raises KicadFileFormatError in case of problems
        """
        library = KicadLibrary(filename)

        # read the s-expression data
        with open(filename) as f:
            # parse s-expr
            try:
                sexpr_data = sexpr.parse_sexp(f.read())
            except ValueError as exc:
                raise KicadFileFormatError(
                    f"Problem while parsing the s-expr file: {exc}"
                ) from None
        sym_list = _get_array(sexpr_data, "symbol", max_level=2)

        # Because of the various file format changes in the development of kicad v6 and v7, we want
        # to ensure that this parser is only used with v6 files. Any other version will most likely
        # not work as expected. So just don't load them at all.
        version = _get_value_of(sexpr_data, "version")
        if str(version) != "20211014":
            raise KicadFileFormatError('Version of symbol file is not "20211014"')

        # itertate over symbol
        for item in sym_list:
            item_type = item.pop(0)
            if item_type != "symbol":
                raise KicadFileFormatError(f"Unexpected token found: {item_type}")
            # retrieving the `partname`, even if formatted as `libname:partname` (legacy format)
            partname = item.pop(0).split(":")[-1]

            # we found a new part, extract the symbol name
            symbol = KicadSymbol(str(partname), libname=filename, filename=filename)

            # extract extends property
            extends = _get_array2(item, "extends")
            if extends:
                symbol.extends = extends[0][1]

            # extract properties
            for prop in _get_array(item, "property"):
                try:
                    # TODO: do not append the new property, if it is None
                    symbol.properties.append(Property.from_sexpr(prop))
                except ValueError as exc:
                    raise KicadFileFormatError(
                        f"Failed to import '{partname}': {exc}"
                    ) from exc

            # get flags
            symbol.in_bom = _get_value_of(item, "in_bom", "no") == "yes"
            symbol.on_board = _get_value_of(item, "on_board", "no") == "yes"
            if _has_value(item, "power"):
                symbol.is_power = True

            # get pin-numbers properties
            pin_numbers_info = _get_array2(item, "pin_numbers")
            if pin_numbers_info:
                if "hide" in pin_numbers_info[0]:
                    symbol.hide_pin_numbers = True

            # get pin-name properties
            pin_names_info = _get_array2(item, "pin_names")
            if pin_names_info:
                if "hide" in pin_names_info[0]:
                    symbol.hide_pin_names = True
                # sometimes the pin_name_offset value does not exist, then use 20mil as default
                symbol.pin_names_offset = _get_value_of(
                    pin_names_info[0], "offset", 0.508
                )

            # get the actual geometry information
            # it is split over units
            subsymbols = _get_array2(item, "symbol")
            for unit_data in subsymbols:
                # we found a new 'subpart' (no clue how to call it properly)
                subpart_type = unit_data.pop(0)
                if subpart_type != "symbol":
                    raise KicadFileFormatError(
                        f"Unexpected token found as 'subsymbol' of {item_type}: {subpart_type}"
                    )
                name = unit_data.pop(0)

                # split the name
                m1 = re.match(r"^" + re.escape(partname) + r"_(\d+?)_(\d+?)$", name)
                if not m1:
                    raise KicadFileFormatError(
                        "Failed to parse subsymbol due to invalid name: {name}"
                    )

                (unit_idx, demorgan_idx) = (m1.group(1), m1.group(2))
                unit_idx = int(unit_idx)
                demorgan_idx = int(demorgan_idx)

                # update the amount of units, alternative-styles (demorgan)
                symbol.unit_count = max(unit_idx, symbol.unit_count)
                symbol.demorgan_count = max(demorgan_idx, symbol.demorgan_count)

                # extract pins and graphical items
                for pin in _get_array(unit_data, "pin"):
                    try:
                        symbol.pins.append(Pin.from_sexpr(pin, unit_idx, demorgan_idx))
                    except ValueError as valexc:
                        raise KicadFileFormatError(
                            f"Failed to parse symbol {partname}: {valexc}"
                        ) from None
                for circle in _get_array(unit_data, "circle"):
                    symbol.circles.append(
                        Circle.from_sexpr(circle, unit_idx, demorgan_idx)
                    )
                for arc in _get_array(unit_data, "arc"):
                    symbol.arcs.append(Arc.from_sexpr(arc, unit_idx, demorgan_idx))
                for rect in _get_array(unit_data, "rectangle"):
                    # symbol.polylines.append(
                    #     Rectangle.from_sexpr(rect, unit, demorgan).as_polyline()
                    # )
                    symbol.rectangles.append(
                        Rectangle.from_sexpr(rect, unit_idx, demorgan_idx)
                    )
                for poly in _get_array(unit_data, "polyline"):
                    symbol.polylines.append(
                        Polyline.from_sexpr(poly, unit_idx, demorgan_idx)
                    )
                for text in _get_array(unit_data, "text"):
                    symbol.texts.append(Text.from_sexpr(text, unit_idx, demorgan_idx))

            # add it to the list of symbols
            library.symbols.append(symbol)

        return library


if __name__ == "__main__":
    if len(sys.argv) >= 2:
        a = KicadLibrary.from_file(sys.argv[1])
        a.generator = "kicad_symbol_editor"
        a.version = "20200908"
        print(a.get_sexpr())
    else:
        print("pass a .kicad_sym file please")
